
/* //////////////////////////////////////////////////// */
/* This file is a part of the BSTools procedure package */
/* written by Przemyslaw Kiciak.                        */
/* //////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <unistd.h>
#include <sys/times.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "pkvaria.h"
#include "pkgeom.h"
#include "pknum.h"
#include "multibs.h"
#define XGLE_OPENGL_OLD
#include "xgledit.h"

#include "light.h"
#include "sphericon.h"
#include "plane.h"
#include "frames.h"

#define ENGLISH
/*#define POLISH*/
/* ///////////////////////////////////////////////////////////////////////// */
#define POSQUEUELENGTH 500

#define MENUWIDTH 90

#define btnQUIT       1
#define intwN         2
#define intwK         3
#define intwH         4
#define swDUAL        5
#define swHALFDUAL    6
#define swENLIGHT     7
#define swROLL1       8
#define swROLL2       9
#define intwSTART    10
#define btnANIMATEFW 11
#define btnANIMATEBK 12
#define btnBK        13
#define btnFW        14
#define slSPEED      15
#define slDELAY      16
#define swFOLLOW1    17
#define swFOLLOW2    18
#define swFOLLOW3    19
#define swANTIALIAS  20
#define swFRAMES     21

#define cmdANIMATE 1

xge_3Dwind     wind;
xge_int_widget intw_sph_n, intw_sph_k, intw_sph_h, intw_start;
int            sph_n, sph_k, sph_h, start = 0;
boolean        dual, halfdual;
xge_widget     *menu;
double         speed = 0.15;
double         delay = 0.5;
double         delta = 0.05;

boolean        rolling = false, rolling1 = false, rolling2 = false,
               antialias = false;
clock_t        t0, t1;
double         clk_tck;
boolean        animate_fw = false, animate_bk = false,
               follow1 = false, follow2 = true, follow3 = false,
               enlight = false;
xge_widget     *sw_follow3;

boolean        frames = false;

/* OpenGL display lists */
#define NLISTS      5
#define SPH_WF      0
#define PLANE_WF    1
#define SPH_SOLID   2
#define PLANE_SOLID 3
#define SHADOW      4
GLuint         dlist;
boolean        listval[NLISTS];

/* ///////////////////////////////////////////////////////////////////////// */
clock_t GetMyTick ( void )
{
  struct tms tt;

  if ( frames )
    return (clock_t)(frame_cnt*(int)(clk_tck/25.0));
  else
    return times ( &tt );
} /*GetMyTick*/

/* ///////////////////////////////////////////////////////////////////////// */
void DrawEnlightedSphericon ( void )
{
  if ( listval[SPH_SOLID] )
    glCallList ( dlist+SPH_SOLID );
  else {
    glNewList ( dlist+SPH_SOLID, GL_COMPILE_AND_EXECUTE );
    DrawSphericonSolid ();
    glEndList ();
    listval[SPH_SOLID] = true;
  }
} /*DrawEnlightedSphericon*/

void DrawEnlightedScene ( int id )
{
  trans3d tr;

  glEnable ( GL_DEPTH_TEST );
  glEnable ( GL_NORMALIZE );
  LightsOn ( &wind.CPos[id] );
  if ( rolling1 ) {
    SetupDropShadowTrans ( 0, &tr );
    if ( wind.CPos[id].position.z >= 0.0 )
      BindPlaneTexture ();
    if ( listval[SHADOW] )
      glCallList ( dlist+SHADOW );
    else {
      glNewList ( dlist+SHADOW, GL_COMPILE_AND_EXECUTE );
      glDisable ( GL_LIGHT0 );
      DrawSphericonShadow ( &sphtrans1, &tr, sph_h >= 0 );
      glEnable ( GL_LIGHT0 );
      glEndList ();
      listval[SHADOW] = true;
    }
    if ( wind.CPos[id].position.z >= 0.0 ) {
      if ( listval[PLANE_SOLID] )
        glCallList ( dlist+PLANE_SOLID );
      else {
        glNewList ( dlist+PLANE_SOLID, GL_COMPILE_AND_EXECUTE );
        DrawThePlaneSolid ();
        glEndList ();
        listval[PLANE_SOLID] = true;
      }
      glBindTexture ( GL_TEXTURE_2D, 0 );
      glDisable ( GL_TEXTURE_2D );
    }
    xgl2e_MultMatrix3d ( &sphtrans1 );
  }
  else if ( rolling2 ) {
    if ( follow3 ) {
      glPushMatrix ();
      xgl2e_MultMatrix3d ( &sphtrans3 );
      DrawEnlightedSphericon ();
      glPopMatrix ();
    }
    else {
      glPushMatrix ();
      xgl2e_MultMatrix3d ( &sphtrans2 );
      DrawEnlightedSphericon ();
      glPopMatrix ();
      xgl2e_MultMatrix3d ( &sphtrans1 );
    }
  }
  DrawEnlightedSphericon ();
  LightsOff ();
  glDisable ( GL_NORMALIZE );
  glDisable ( GL_DEPTH_TEST );
} /*DrawEnlightedScene*/

void DrawWireframeSphericon ( void )
{
  if ( listval[SPH_WF])
    glCallList ( dlist+SPH_WF );
  else {
    glNewList ( dlist+SPH_WF, GL_COMPILE_AND_EXECUTE );
    DrawSphericonWF ();
    glEndList ();
    listval[SPH_WF] = true;
  }
} /*DrawWireframeSphericon*/

void DrawWireframeScene ( void )
{
  glEnable ( GL_DEPTH_TEST );
  if ( rolling1 ) {
    if ( listval[PLANE_WF] )
      glCallList ( dlist+PLANE_WF );
    else {
      glNewList ( dlist+PLANE_WF, GL_COMPILE_AND_EXECUTE );
      DrawThePlaneWF ();
      DrawHCFrame ();
      glEndList ();
      listval[PLANE_WF] = true;
    }
    xgl2e_MultMatrix3d ( &sphtrans1 );
  }
  else if ( rolling2 ) {
    if ( follow3 ) {
      glPushMatrix ();
      xgl2e_MultMatrix3d ( &sphtrans3 );
      DrawWireframeSphericon ();
      glPopMatrix ();
    }
    else {
      glPushMatrix ();
      xgl2e_MultMatrix3d ( &sphtrans2 );
      DrawWireframeSphericon ();
      glPopMatrix ();
      xgl2e_MultMatrix3d ( &sphtrans1 );
    }
  }
  DrawWireframeSphericon ();
  glDisable ( GL_DEPTH_TEST );
} /*DrawWireframeScene*/

void RysujOkno ( xge_widget *er, boolean onscreen )
{
#define NTR 9
  int id, i;
  static double pshx[NTR] = {-0.4, -0.4, -0.4, 0.0, 0.0, 0.0, 0.4, 0.4, 0.4 };
  static double pshy[NTR] = {-0.4, 0.0, 0.4, -0.4, 0.0, 0.4, -0.4, 0.0, 0.4 };

  XSync ( xgedisplay, false );
  id = er->id & 0x03;
  if ( antialias && _xgle_accum ) {
    glClear ( GL_ACCUM_BUFFER_BIT );
    for ( i = 0; i < NTR; i++ ) {
      xgle_DrawGeomWinBackground ( er, GL_DEPTH_BUFFER_BIT );
      xgl2e_SetIdentMapping ( er );
      xgl2e_SetGLaccCamerad ( &wind.CPos[id], pshx[i], pshy[i],
                              0.0, 0.0, wind.CPos[id].zmin );
      if ( enlight )
        DrawEnlightedScene ( id );
      else
        DrawWireframeScene ();
      glAccum ( i == 0 ? GL_LOAD : GL_ACCUM, 1.0/(double)NTR );
    }
    glAccum ( GL_RETURN, 1.0 );
  }
  else {
    xgle_DrawGeomWinBackground ( er, GL_DEPTH_BUFFER_BIT );
    xgl2e_SetIdentMapping ( er );
    xgl2e_SetGLCamerad ( &wind.CPos[id] );
    if ( enlight )
      DrawEnlightedScene ( id );
    else
      DrawWireframeScene ();
  }
  glFinish ();
  xgleCopyGLRect ( er->w, er->h, er->x, er->y );

  xge_DrawGeomWinSelectionRect ( er, &wind.selection_rect );
  xge_3DwinfDrawGeomWidgets ( er );
  xge_DrawGeomWinFrame ( er, onscreen );
} /*RysujOkno*/

/* ///////////////////////////////////////////////////////////////////////// */
typedef struct {
    time_t  tck;
    point3d pos;
  } posq_item;

pkv_queue  *posqueue;
double     incr;

void GetCurrentPos ( posq_item *pqi )
{
  if ( pkv_QueueFull ( posqueue ) )
    pkv_QueueRemoveFirst ( posqueue, NULL );
  if ( follow3 )
    SetPoint3d ( &pqi->pos, 0.0, 0.0, 0.0 );
  else
    GetSphCentre ( &pqi->pos );
  if ( rolling2 )
    pqi->pos.z = 0.0;
  pqi->tck = GetMyTick ();
  pkv_QueueInsert ( posqueue, (void*)pqi );
} /*GetCurrentPos*/

void FollowByCamera ( point3d *newcp )
{
  vector3d t;

  SubtractPoints3d ( newcp, &wind.CPos[3].g_centre, &t );
  AddVector3d ( &wind.CPos[3].position, &t, &wind.CPos[3].position );
  wind.CPos[3].g_centre = *newcp;
  CameraSetMappingd ( &wind.CPos[3] );
} /*FollowByCamera*/

void ResetCameraPos ( void )
{
  SubtractPoints3d ( &wind.CPos[3].position, &wind.CPos[3].g_centre,
                     &wind.CPos[3].position );
  SetPoint3d ( &wind.CPos[3].g_centre, 0.0, 0.0, 0.0 );
  CameraSetMappingd ( &wind.CPos[3] );
} /*ResetCameraPos*/

boolean MoveIt ( void )
{
  posq_item pqi0, pqi1, *pqi;
  clock_t   tck;
  int       i, n;
  point3d   pq;
  
  if ( animate_fw || animate_bk ) {
    listval[PLANE_SOLID] = listval[SHADOW] = listval[PLANE_WF] = false;
    t1 = GetMyTick ();
    incr = (0.5+4.5*speed)*(double)(t1-t0)/clk_tck;
    if ( incr > 0.1 )
      incr = 0.1;
    if ( animate_bk )
      incr = -incr;
    t0 = t1;
    IncreaseSphTrans ( incr );
    GetCurrentPos ( &pqi1 );
    pqi0 = pqi1;
/*
if ( delay > 0.0 )
  printf ( "fps = %d\n", (int)(pkv_QueueGetNElem ( posqueue )/delay) );
*/
    if ( follow1 || follow2 || follow3 ) {
      pkv_QueueGetFirst ( posqueue, &pqi0 );
      tck = pqi1.tck-pqi0.tck;
      while ( tck > delay*clk_tck && pkv_QueueGetNElem ( posqueue ) > 1 ) {
        pkv_QueueRemoveFirst ( posqueue, NULL );
        pkv_QueueGetFirst ( posqueue, &pqi0 );
        tck = pqi1.tck-pqi0.tck;
      }
      if ( follow2 || follow3 ) {
        n = pkv_QueueGetNElem ( posqueue );
        pq = pqi0.pos;
        for ( i = 1; i < n; i++ ) {
          pqi = (posq_item*)pkv_QueueGetNth ( posqueue, i );
          AddVector3d ( &pq, &pqi->pos, &pq );
        }
        MultVector3d ( 1.0/(double)n, &pq, &pq );
        FollowByCamera ( &pq );
      }
      else
        FollowByCamera ( &pqi0.pos );
    }
    SetPlaneCentre ( wind.CPos[3].g_centre.x, wind.CPos[3].g_centre.y );
    xge_Redraw ();
    xge_PostIdleCommand ( cmdANIMATE, 0, 0 );
    if ( frames )
      frames = WriteFrame ( wind.fww.er->w, wind.fww.er->h,
                            wind.fww.er->x, wind.fww.er->y,
                            sph_n, sph_k, dual, halfdual );
  }
  return 1;
} /*MoveIt*/

/* ///////////////////////////////////////////////////////////////////////// */
int CallBack ( xge_widget *er, int msg, int key, short x, short y )
{
  posq_item pqi;

  if ( er ) {
/*    printf ( "%d, %d, %d, %d, %d\n", er->id, msg, key, x, y ); */
    switch ( msg ) {
case xgemsg_BUTTON_COMMAND:
      switch ( er->id ) {
  case btnQUIT:
        xge_done = 1;
        break;
  case btnBK:
        if ( !rolling ) {
          rolling = rolling1 = true;
          rolling2 = false;
          InitSphTrans ( start );
          pkv_ResetQueue ( posqueue );
          GetCurrentPos ( &pqi );
        }
        animate_fw = animate_bk = false;
        listval[SHADOW] = false;
        IncreaseSphTrans ( -0.05 );
        GetCurrentPos ( &pqi );
        xge_Redraw ();
        break;
  case btnFW:
        if ( !rolling ) {
          rolling = rolling1 = true;
          rolling2 = false;
          InitSphTrans ( start );
          pkv_ResetQueue ( posqueue );
          GetCurrentPos ( &pqi );
        }
        animate_fw = animate_bk = false;
        listval[SHADOW] = false;
        IncreaseSphTrans ( 0.05 );
        GetCurrentPos ( &pqi );
        xge_Redraw ();
        break;
  case btnANIMATEFW:
        if ( animate_fw ) {
          animate_fw = false;
          xge_Redraw ();
        }
        else {
          animate_fw = true;  animate_bk = false;
          if ( !rolling ) {
            rolling = rolling1 = true;
            rolling2 = false;
            InitSphTrans ( start );
            pkv_ResetQueue ( posqueue );
            GetCurrentPos ( &pqi );
          }
          t0 = GetMyTick ();
          xge_PostIdleCommand ( cmdANIMATE, 0, 0 );
          xge_Redraw ();
          if ( frames )
            frames = WriteFrame ( wind.fww.er->w, wind.fww.er->h,
                                  wind.fww.er->x, wind.fww.er->y,
                                  sph_n, sph_k, dual, halfdual );
        }
        break;
  case btnANIMATEBK:
        if ( animate_bk ) {
          animate_bk = false;
          xge_Redraw ();
        }
        else {
          animate_bk = true;  animate_fw = false;
          if ( !rolling ) {
            rolling = rolling1 = true;
            rolling2 = false;
            InitSphTrans ( start );
            pkv_ResetQueue ( posqueue );
            GetCurrentPos ( &pqi );
          }
          t0 = GetMyTick ();
          xge_PostIdleCommand ( cmdANIMATE, 0, 0 );
          xge_Redraw ();
        }
        break;
  default:
        break;
      }
      break;

case xgemsg_SWITCH_COMMAND:
      switch ( er->id ) {
  case swDUAL:
        if ( dual ) {
          if ( sph_n & 0x01 ) {
            dual = false;
            menu->redraw ( menu, true );
          }
          else {
            halfdual = false;
            goto generate;
          }
        }
        else
          goto generate;
        break;
  case swHALFDUAL:
        if ( halfdual ) {
          if ( sph_n & 0x01 ) {
            halfdual = false;
            menu->redraw ( menu, true );
          }
          else {
            dual = false;
            goto generate;
          }
        }
        else
          goto generate;
        break;
  case swROLL1:
        if ( rolling1 )
          rolling2 = follow3 = false;
        else
          animate_fw = animate_bk = false;
        sw_follow3->state = xgestate_SWITCH_INACTIVE;
        if ( (rolling = rolling1) ) {
          SetPlaneCentre ( 0.0, 0.0 );
          InitSphTrans ( start );
          xge_Redraw ();
        }
        else
          goto roll_off;
        break;
  case swROLL2:
        if ( rolling2 ) {
          sw_follow3->state = xgestate_SWITCH_DEFAULT;
          rolling1 = false;
        }
        else {
          sw_follow3->state = xgestate_SWITCH_INACTIVE;
          animate_fw = animate_bk = false;
        }
        if ( (rolling = rolling2) ) {
          SetPlaneCentre ( 0.0, 0.0 );
          InitSphTrans ( start );
          xge_Redraw ();
        }
        else {
roll_off:
          IdentSphTrans ();
          ResetCameraPos ();
          xge_Redraw ();
        }
        break;
  case swENLIGHT:
        xge_Redraw ();
        break;
  case swANTIALIAS:
        if ( !_xgle_accum )
          antialias = false;
        else
          xge_Redraw ();
        break;
  case swFOLLOW1:
        if ( follow1 )
          follow2 = follow3 = false;
        xge_Redraw ();
        break;
  case swFOLLOW2:
        if ( follow2 )
          follow1 = follow3 = false;
        xge_Redraw ();
        break;
  case swFOLLOW3:
        if ( follow3 )
          follow1 = follow2 = false;
        xge_Redraw ();
        break;
  default:
        break;
      }
      break;

case xgemsg_SLIDEBAR_COMMAND:
      switch ( er->id ) {
  case slSPEED:
        break;
  case slDELAY:
        break;
  default:
        break;
      }
      break;

case xgemsg_INT_WIDGET_COMMAND:
      switch ( er->id ) {
  case intwN:
        if ( key & 0x01 )
          dual = halfdual = false;
        sph_n = key;
        if ( sph_k >= sph_n )
          sph_k = sph_n-1;
        if ( start >= sph_n )
          start = sph_n-1;
        goto generate;
  case intwH:
        if ( key == 0 )
          sph_h = -2;
        else if ( key == -1 )
          sph_h = 1;
        else
          sph_h = key;
        goto generate;
  case intwK:
        if ( key < sph_n )
          sph_k = key;
generate:
        rolling = rolling1 = rolling2 = animate_fw = animate_bk = false;
        GenSphericon ( sph_n, sph_k, sph_h, dual, halfdual );
        if ( start >= nhcones )
          start = nhcones-1;
        ResetCameraPos ();
        memset ( listval, false, NLISTS );
        xge_Redraw ();
        InitFrameSequence ();
        break;
  case intwSTART:
        if ( key >= 0 && key < nhcones ) {
          start = key;
          InitSphTrans ( start );
          if ( rolling )
            xge_Redraw ();
          return 1;
        }
        break;
  default:
        break;
      }
      break;

case xgemsg_3DWIN_RESIZE:
case xgemsg_3DWIN_PROJCHANGE:
      break;

default:
      break;
    }
  }
  else {
/*    printf ( "%d, %d, %d, %d\n", msg, key, x, y ); */
    switch ( msg ) {
case xgemsg_KEY:
      switch ( key ) {
  case 'm':
        if ( xge_current_width != xge_WIDTH ||
             xge_current_height != xge_HEIGHT )
          xgeResizeWindow ( xge_WIDTH, xge_HEIGHT );
        else
          xgeResizeWindow ( 800+MENUWIDTH, 600 );
        break;
  case 'M':
        if ( xge_current_width != xge_MAX_WIDTH ||
             xge_current_height != xge_MAX_HEIGHT )
          xgeResizeWindow ( xge_MAX_WIDTH, xge_MAX_HEIGHT );
        else
          xgeResizeWindow ( 1024+MENUWIDTH, 768 );
        break;
  case 'Q': case 'q':
        xge_done = 1;
        break;
  default:
        break;
      }
      break;

case xgemsg_RESIZE:
      wind.fww.er->msgproc ( wind.fww.er, xgemsg_RESIZE, 0,
                   (short)(xge_current_width-MENUWIDTH), xge_current_height );
      menu->msgproc ( menu, xgemsg_RESIZE, 0, MENUWIDTH, xge_current_height );
      xge_Redraw ();
      break;

case xgemsg_IDLE_COMMAND:
    switch ( key ) {
  case cmdANIMATE: 
        return MoveIt ();
      default:
        return 0;
      }

default:
      break;
    }
  }
  return 0;
} /*CallBack*/

/* ///////////////////////////////////////////////////////////////////////// */
#ifdef ENGLISH
static char txtQuit[]      = "Quit";
static char txtDual[]      = "dual";
static char txtHalfdual[]  = "halfdual";
static char txtN[]         = "n";
static char txtK[]         = "k";
static char txtH[]         = "h";
static char txtEnlight[]   = "enlight";
static char txtRoll1[]     = "roll 1";
static char txtRoll2[]     = "roll 2";
static char txtLeft[]      = " <-";
static char txtRight[]     = " ->";
static char txtStart[]     = "begin at";
static char txtFollow1[]   = "follow 1";
static char txtFollow2[]   = "follow 2";
static char txtFollow3[]   = "follow 3";
static char txtAntialias[] = "aa";
static char txtFrames[]    = "frames";
#endif
#ifdef POLISH
static char txtQuit[]      = "Koniec";
static char txtDual[]      = "dualny";
static char txtHalfdual[]  = "poldualny";
static char txtN[]         = "n";
static char txtK[]         = "k";
static char txtH[]         = "h";
static char txtEnlight[]   = "oswietl";
static char txtRoll1[]     = "turlaj 1";
static char txtRoll2[]     = "turlaj 2";
static char txtLeft[]      = " <-";
static char txtRight[]     = " ->";
static char txtStart[]     = "zacznij od";
static char txtFollow1[]   = "ganiaj 1";
static char txtFollow2[]   = "ganiaj 2";
static char txtAntialias[] = "aa";
static char txtFrames[]    = "klatki";
#endif

static char *InfoMsg[5] =
  { "Sphericons demonstration program",
    "",
    "This program is a part of the BSTools package.",
    "(C) Copyright by Przemyslaw Kiciak, 2015",
    NULL };

boolean FourWWMsgFilter ( xge_widget *er, int msg, int key, short x, short y )
{
  if ( msg == xgemsg_KEY ) {
    if ( key == 's' || key == 'S' )
      return false;
  }
  return xge_FourWWMsg ( er, msg, key, x, y );
} /*FourWWMsgFilter*/

void MyDrawButton ( xge_widget *er, boolean onscreen )
{
  XPoint p[3];
  short  xc, yc;

  xge_DrawButton ( er, false );
  xgeSetForeground ( xgec_White );
  xc = er->x + er->w/2;
  yc = er->y + er->h/2;
  switch ( er->id ) {
case btnANIMATEBK:
    if ( animate_bk ) {
      xgeFillRectangle ( 2, 10, xc-4, yc-5 );
      xgeFillRectangle ( 2, 10, xc+2, yc-5 );
    }
    else {
      p[0].x = p[2].x = xc+5;  p[1].x = xc-5;
      p[0].y = yc-5;  p[1].y = yc;  p[2].y = yc+5;
      xgeFillPolygon ( Convex, 3, p );
    }
    break;
case btnANIMATEFW:
    if ( animate_fw ) {
      xgeFillRectangle ( 2, 10, xc-4, yc-5 );
      xgeFillRectangle ( 2, 10, xc+2, yc-5 );
    }
    else {
      p[0].x = p[2].x = xc-5;  p[1].x = xc+5;
      p[0].y = yc-5;  p[1].y = yc;  p[2].y = yc+5;
      xgeFillPolygon ( Convex, 3, p );
    }
    break;
default:
    break;
  }
  if ( onscreen )
    xgeCopyRectOnScreen ( er->w, er->h, er->x, er->y );
} /*MyDrawButton*/

void init_edwin ( void )
{
  xge_widget *rp, *edr;

  pkv_InitScratchMem ( 10485760 );
  posqueue = pkv_InitQueue ( POSQUEUELENGTH, sizeof(posq_item) );
  if ( !posqueue )
    exit ( 1 );
  sph_n = 4;
  sph_k = 1;
  sph_h = 1;
  dual = halfdual = false;
  GenSphericon ( sph_n, sph_k, sph_h, dual, halfdual );
  rp = xge_NewButton ( 0, NULL, btnQUIT, MENUWIDTH, 18, 0, 0, txtQuit );
  rp = xge_NewIntWidget ( 0, rp, intwN, 28, 18, 0, 20, 3, MAXCONES,
                          &intw_sph_n, txtN, &sph_n );
  rp = xge_NewIntWidget ( 0, rp, intwK, 28, 18, 31, 20, 0, MAXCONES/2,
                          &intw_sph_k, txtK, &sph_k );
  rp = xge_NewIntWidget ( 0, rp, intwH, 28, 18, 62, 20, -MAXH, MAXH,
                          &intw_sph_h, txtH, &sph_h );
  rp = xge_NewSwitch ( 0, rp, swDUAL, MENUWIDTH, 16, 0, 40, txtDual, &dual );
  rp = xge_NewSwitch ( 0, rp, swHALFDUAL, MENUWIDTH, 16, 0, 60, txtHalfdual, &halfdual );
  rp = xge_NewSwitch ( 0, rp, swENLIGHT, MENUWIDTH, 16, 0, 80, txtEnlight, &enlight );
  rp = xge_NewSwitch ( 0, rp, swROLL1, MENUWIDTH, 16, 0, 100, txtRoll1, &rolling1 );
  rp = xge_NewSwitch ( 0, rp, swROLL2, MENUWIDTH, 16, 0, 120, txtRoll2, &rolling2 );
  rp = xge_NewIntWidget ( 0, rp, intwSTART, MENUWIDTH, 18, 0, 140, 0, MAXH*MAXCONES,
                          &intw_start, txtStart, &start );
  rp = xge_NewButton ( 0, rp, btnBK, 28, 18, 0, 160, txtLeft );
  rp = xge_NewButton ( 0, rp, btnFW, 28, 18, 30, 160, txtRight );
  rp = xge_NewButton ( 0, rp, btnANIMATEBK, 28, 18, 0, 180, NULL );
  rp->redraw = MyDrawButton;
  rp = xge_NewButton ( 0, rp, btnANIMATEFW, 28, 18, 30, 180, NULL );
  rp->redraw = MyDrawButton;
  rp = xge_NewSlidebard ( 0, rp, slSPEED, MENUWIDTH, 10, 0, 204, &speed );
  rp = xge_NewSlidebard ( 0, rp, slDELAY, MENUWIDTH, 10, 0, 220, &delay );
  rp = xge_NewSwitch ( 0, rp, swFOLLOW1, MENUWIDTH, 16, 0, 240, txtFollow1, &follow1 );
  rp = xge_NewSwitch ( 0, rp, swFOLLOW2, MENUWIDTH, 16, 0, 260, txtFollow2, &follow2 );
  sw_follow3 = rp =
       xge_NewSwitch ( 0, rp, swFOLLOW3, MENUWIDTH, 16, 0, 280, txtFollow3, &follow3 );
  sw_follow3->state = xgestate_SWITCH_INACTIVE;
  rp = xge_NewSwitch ( 0, rp, swANTIALIAS, MENUWIDTH, 16, 0, 300,
                       txtAntialias, &antialias );
  if ( !_xgle_accum )
    rp->state = xgestate_SWITCH_INACTIVE;
  antialias = _xgle_accum;
  rp = xge_NewSwitch ( 0, rp, swFRAMES, MENUWIDTH, 16, 0, 320, txtFrames, &frames );
  menu = xge_NewMenu ( 0, NULL, 1, MENUWIDTH, xge_HEIGHT, 0, 0, rp );
  edr = xge_New3Dwind ( 0, menu, 0, xge_WIDTH-MENUWIDTH, xge_HEIGHT, MENUWIDTH, 0,
                        &wind, RysujOkno, RysujOkno );
  xge_SetWinEdRect ( edr );
  CameraSetDepthRanged ( &wind.CPos[0], -10.0,  10.0 );
  CameraSetDepthRanged ( &wind.CPos[1], -10.0,  10.0 );
  CameraSetDepthRanged ( &wind.CPos[2], -10.0,  10.0 );
  CameraSetDepthRanged ( &wind.CPos[3],   1.0, 100.0 );
  wind.fww.zoomwin = 3;
  wind.fww.er->msgproc ( wind.fww.er, xgemsg_RESIZE, 0,
                         wind.fww.er->w, wind.fww.er->h );
  wind.fww.er->msgproc = FourWWMsgFilter;
  clk_tck = (double)sysconf ( _SC_CLK_TCK );
  dlist = glGenLists ( NLISTS );
  if ( !dlist )
    exit ( 1 );
  InitXImage ();
  InitPlaneTexture ();
  SetupLights ();
  xge_Redraw ();
  xge_DisplayInfoMessage ( InfoMsg, -1 );
} /*init_edwin*/

void destroy_edwin ( void )
{
  PKV_FREE ( posqueue );
  pkv_DestroyScratchMem ();
} /*destroy_edwin*/

int main ( int argc, char *argv[] )
{
  if ( !xgle_Init ( argc, argv, CallBack, NULL, 2, 1,
                    XGLE_DEPTH_BUFFER, XGLE_ACCUM_BUFFER ) )
    exit ( 1 );
  init_edwin ();
  xge_MessageLoop ();
  destroy_edwin ();  
  xgle_Cleanup ();  
  exit ( 0 );
} /*main*/   

