
/* //////////////////////////////////////////////////// */
/* This file is a part of the BSTools procedure package */
/* written by Przemyslaw Kiciak.                        */
/* //////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>   
#include <malloc.h>
#include <string.h>

#include <GL/gl.h>

#include "pkvaria.h"
#include "pkgeom.h"
#include "pknum.h"  
#include "multibs.h"
#include "camera.h"
#include "convh.h"

#include "light.h"
#include "sphericon.h"

#define MAXLEVELWF    3
#define MAXLEVELSOLID 4

/* ///////////////////////////////////////////////////////////////////////// */
static void rDrawBezPatchWF ( point4d *cp, int level )
{
  point4d cq[6];

  pkv_Selectd ( 3, 4, 8, 4, cp, cq );
  pkv_Selectd ( 3, 4, 8, 4, &cp[1], &cq[3] );
  mbs_BezierStandardC3Rd ( 2, cq );
  mbs_BezierStandardC3Rd ( 2, &cq[3] );
  pkv_Selectd ( 3, 4, 4, 8, cq, cp );  
  pkv_Selectd ( 3, 4, 4, 8, &cq[3], &cp[1] );
  mbs_multiBisectBezCurvesd ( 2, 1, 8, 0, &cp[0].x, &cq[0].x );
  if ( level == MAXLEVELWF ) {
    glVertex4dv ( &cp[4].x );
    glVertex4dv ( &cp[5].x );
  }
  else {
    rDrawBezPatchWF ( cq, level+1 );
    rDrawBezPatchWF ( cp, level+1 );
  }
} /*rDrawBezPatchWF*/

static void rDrawBezPatchWF1 ( point4d *cp, int level )
{
  point4d cq[3];

  mbs_BezierStandardC3Rd ( 2, cp );
  mbs_BisectBC4d ( 2, &cp[0].x, &cq[0].x );
  if ( level == MAXLEVELWF )
    glVertex4dv ( &cp[0].x );
  else {
    rDrawBezPatchWF1 ( cq, level+1 );
    glVertex4dv ( &cp[0].x );
    rDrawBezPatchWF1 ( cp, level+1 );
  }
} /*rDrawBezPatchWF1*/

static void DrawBezPatchWF ( point4d *cp )
{
  point4d cc[6];

  pkv_Selectd ( 3, 4, 8, 4, &cp[0].x, &cc[0].x );
  pkv_Selectd ( 3, 4, 8, 4, &cp[1].x, &cc[3].x );
  glBegin ( GL_LINES );
    glVertex4dv ( &cp[0].x );
    glVertex4dv ( &cp[1].x );
    rDrawBezPatchWF ( cp, 0 );
    glVertex4dv ( &cp[4].x );
    glVertex4dv ( &cp[5].x );
  glEnd ();
  glBegin ( GL_LINE_STRIP );
    glVertex4dv ( &cc[0].x );
    rDrawBezPatchWF1 ( cc, 0 );
    glVertex4dv ( &cc[2].x );
  glEnd ();
  glBegin ( GL_LINE_STRIP );
    glVertex4dv ( &cc[3].x );
    rDrawBezPatchWF1 ( &cc[3], 0 );
    glVertex4dv ( &cc[5].x );
  glEnd ();
} /*DrawBezPatchWF*/

void DrawSphericonWF ( void )
{
  int     i;
  point4d cq[6];

  glColor3f ( 0.0, 1.0, 1.0 );
  for ( i = 0; i < nhcones; i++ ) {
    memcpy ( cq, hc[i].qcp0, 6*sizeof(point4d) );
    DrawBezPatchWF ( cq );
    memcpy ( cq, hc[i].qcp1, 6*sizeof(point4d) );
    DrawBezPatchWF ( cq );
    if ( i == 3 )
      glColor3f ( 1.0, 1.0, 0.0 );
    else
      glColor3f ( 0.0, 1.0, 0.0 );
  }
} /*DrawSphericonWF*/

/* ///////////////////////////////////////////////////////////////////////// */
static void rDrawBezPatchSolid ( point4d *cp, int level, double *ws )
{
  point4d  cq[6];
  point3d  pp;
  vector3d nv;

  pkv_Selectd ( 3, 4, 8, 4, cp, cq );
  pkv_Selectd ( 3, 4, 8, 4, &cp[1], &cq[3] );
  mbs_BezierStandardC3Rd ( 2, cq );
  mbs_BezierStandardC3Rd ( 2, &cq[3] );
  pkv_Selectd ( 3, 4, 4, 8, cq, cp );  
  pkv_Selectd ( 3, 4, 4, 8, &cq[3], &cp[1] );
  mbs_multiBisectBezCurvesd ( 2, 1, 8, 0, &cp[0].x, &cq[0].x );
  if ( level == MAXLEVELSOLID ) {
    _mbs_BCHornerNvP3Rd ( 2, 1, cp, 1.0, 0.5, &pp, &nv, ws );
    glNormal3dv ( &nv.x );
    glVertex4dv ( &cp[4].x );
    glVertex4dv ( &cp[5].x );
  }
  else {
    rDrawBezPatchSolid ( cq, level+1, ws );
    rDrawBezPatchSolid ( cp, level+1, ws );
  }
} /*rDrawBezPatchSolid*/

static void DrawBezPatchSolid ( point4d *cp )
{
  point4d  cc[6];
  point3d  pp;
  vector3d nv;
  double   ws[40];

  pkv_Selectd ( 3, 4, 8, 4, &cp[0].x, &cc[0].x );
  pkv_Selectd ( 3, 4, 8, 4, &cp[1].x, &cc[3].x );
  glBegin ( GL_QUAD_STRIP );
    _mbs_BCHornerNvP3Rd ( 2, 1, cp, 0.0, 0.5, &pp, &nv, ws );
    glNormal3dv ( &nv.x );
    glVertex4dv ( &cp[0].x );
    glVertex4dv ( &cp[1].x );
    rDrawBezPatchSolid ( cp, 0, ws );
    _mbs_BCHornerNvP3Rd ( 2, 1, cp, 1.0, 0.5, &pp, &nv, ws );
    glNormal3dv ( &nv.x );
    glVertex4dv ( &cp[4].x );
    glVertex4dv ( &cp[5].x );
  glEnd ();
} /*DrawBezPatchSolid*/

static void _DrawSphericonSolid ( void )
{
  point4d cq[6];
  int     i;

  for ( i = 0; i < nhcones; i++ ) {
    memcpy ( cq, hc[i].qcp0, 6*sizeof(point4d) );
    DrawBezPatchSolid ( cq );
    memcpy ( cq, hc[i].qcp1, 6*sizeof(point4d) );
    DrawBezPatchSolid ( cq );
  }
} /*_DrawSphericonSolid*/

void DrawSphericonSolid ( void )
{
  GLfloat colour[3] = { 1.0, 1.0, 0.0 };
  GLfloat specular[4] = { 1.0, 1.0, 1.0, 1.0 };
  vector3d nv = {0.0,0.0,1.0};

  glMaterialfv ( GL_FRONT_AND_BACK, GL_SPECULAR, specular );
  glMaterialf ( GL_FRONT_AND_BACK, GL_SHININESS, 100.0 );
  glColorMaterial ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
  glEnable ( GL_COLOR_MATERIAL );
  glColor3fv ( colour );
  glNormal3dv ( &nv.x );
  _DrawSphericonSolid ();
  glDisable ( GL_COLOR_MATERIAL );
} /*DrawSphericonSolid*/

/* ///////////////////////////////////////////////////////////////////////// */
static void rDrawBezPatchShadow ( point4d *cp, int level, trans3d *t,
                                  point2d *spsh, int *nsp )
{
  point4d  cq[6], q;

  pkv_Selectd ( 3, 4, 8, 4, cp, cq );
  pkv_Selectd ( 3, 4, 8, 4, &cp[1], &cq[3] );
  mbs_BezierStandardC3Rd ( 2, cq );
  mbs_BezierStandardC3Rd ( 2, &cq[3] );
  pkv_Selectd ( 3, 4, 4, 8, cq, cp );  
  pkv_Selectd ( 3, 4, 4, 8, &cq[3], &cp[1] );
  mbs_multiBisectBezCurvesd ( 2, 1, 8, 0, &cp[0].x, &cq[0].x );
  if ( level == MAXLEVELSOLID ) {
    Trans3Point4d ( t, &cp[4], &q );
    SetPoint2d ( &spsh[(*nsp)++], q.x/q.w, q.y/q.w );
    Trans3Point4d ( t, &cp[5], &q );
    SetPoint2d ( &spsh[(*nsp)++], q.x/q.w, q.y/q.w );
  }
  else {
    rDrawBezPatchShadow ( cq, level+1, t, spsh, nsp );
    rDrawBezPatchShadow ( cp, level+1, t, spsh, nsp );
  }
} /*rDrawBezPatchShadow*/

static void DrawBezPatchShadow ( point4d *cp, trans3d *t,
                                 point2d *spsh, int *nsp )
{
  point4d  cc[6], q;

  pkv_Selectd ( 3, 4, 8, 4, &cp[0].x, &cc[0].x );
  pkv_Selectd ( 3, 4, 8, 4, &cp[1].x, &cc[3].x );
  Trans3Point4d ( t, &cp[0], &q );
  SetPoint2d ( &spsh[(*nsp)++], q.x/q.w, q.y/q.w );
  Trans3Point4d ( t, &cp[1], &q );
  SetPoint2d ( &spsh[(*nsp)++], q.x/q.w, q.y/q.w );
  rDrawBezPatchShadow ( cp, 0, t, spsh, nsp );
  Trans3Point4d ( t, &cp[4], &q );
  SetPoint2d ( &spsh[(*nsp)++], q.x/q.w, q.y/q.w );
  Trans3Point4d ( t, &cp[5], &q );
  SetPoint2d ( &spsh[(*nsp)++], q.x/q.w, q.y/q.w );
} /*DrawBezPatchShadow*/

static boolean _DrawSphericonShadow ( trans3d *t, point2d *spsh, int *nsp )
{
  point4d cq[6];
  int     i;

  *nsp = 0;
  for ( i = 0; i < nhcones; i++ ) {
    memcpy ( cq, hc[i].qcp0, 6*sizeof(point4d) );
    DrawBezPatchShadow ( cq, t, spsh, nsp );
    memcpy ( cq, hc[i].qcp1, 6*sizeof(point4d) );
    DrawBezPatchShadow ( cq, t, spsh, nsp );
  }
  if ( *nsp > 3 )
    return FindConvexHull2d ( nsp, spsh );
  else
    return false;
} /*_DrawSphericonShadow*/

void DrawSphericonShadow ( trans3d *tr, trans3d *trsh, boolean convex )
{
#define MAXSHPOINTS (4*MAXH*MAXCONES*128)
  trans3d t;
  GLfloat colour[3] = { 0.5, 0.5, 0.5 };
  GLfloat shadow[4] = { 0.0, 0.0, 0.0, 1.0 };
  point2d *spsh;
  int     nsp, i, j;
  point4d cq[6];

  spsh = (point2d*)pkv_GetScratchMem ( MAXSHPOINTS*sizeof(point2d) );
  if ( !spsh )
    return;
  CompTrans3d ( &t, trsh, tr );
  glMaterialf ( GL_FRONT_AND_BACK, GL_SHININESS, 10.0 );
  glMaterialfv ( GL_FRONT_AND_BACK, GL_DIFFUSE, shadow );
  glMaterialfv ( GL_FRONT_AND_BACK, GL_SPECULAR, shadow );
  glMaterialfv ( GL_FRONT_AND_BACK, GL_AMBIENT, colour );
  glEnable ( GL_COLOR_MATERIAL );
  glColor3fv ( colour );
  glNormal3d ( 0.0, 0.0, 1.0 );
  if ( convex ) {
    if ( _DrawSphericonShadow ( &t, spsh, &nsp  ) ) {
      glBegin ( GL_POLYGON );
        for ( i = 0; i < nsp; i++ )
          glVertex2dv ( &spsh[i].x );
      glEnd ();
    }
  }
  else {
    for ( i = 0; i < nhcones; i++ ) {
      nsp = 0;
      memcpy ( cq, hc[i].qcp0, 6*sizeof(point4d) );
      DrawBezPatchShadow ( cq, &t, spsh, &nsp );
      memcpy ( cq, hc[i].qcp1, 6*sizeof(point4d) );
      DrawBezPatchShadow ( cq, &t, spsh, &nsp );
      if ( nsp > 3 ) {
        if ( FindConvexHull2d ( &nsp, spsh ) ) {
          glBegin ( GL_POLYGON );
            for ( j = 0; j < nsp; j++ )
              glVertex2dv ( &spsh[j].x );
          glEnd ();
        }
      }
    }
  }
  glDisable ( GL_COLOR_MATERIAL );
  pkv_SetScratchMemTop ( (void*)spsh );
} /*DrawSphericonShadow*/

/* ///////////////////////////////////////////////////////////////////////// */
void DrawHCFrame ( void )
{
  point3d c;

  glPointSize ( 5.0 );
  glBegin ( GL_POINTS );
    glColor3f ( 0.0, 1.0, 0.0 );
    glVertex3dv ( &hcframe[0].x );
    glColor3f ( 0.0, 0.8, 0.2 );
    if ( hc[rollhc].cyl ) {
      AddVector3d ( &hcframe[0], &hcframe[1], &c );
      glVertex3dv ( &c.x );
    }
    else
      glVertex3dv ( &hcframe[1].x );
    glColor3f ( 0.0, 0.6, 0.4 );
    glVertex3dv ( &hcframe[2].x );
    glColor3f ( 0.0, 0.4, 0.6 );
    glVertex3dv ( &hcframe[3].x );
  glEnd ();
} /*DrawHCFrame*/

