#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <sys/times.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "pkvaria.h"
#include "xgedit.h"

#include "suHD.h"

/* ////////////////////////////////////////////////////////////////////////// */
char permut[24][4];
char hpermut[16] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15};
int  par[10] = {0,0,0,0,0,0,0,0,0,0};

char           su[16][16], suuser[16][16];
char           sustatus[16][16];
unsigned short suprompts[16][16];
unsigned char  hintpermut[256], hpinv[256];

const char hd[16] = {'0','1','2','3','4','5','6','7',
                     '8','9','A','B','C','D','E','F'};
int  cell_size = CELL_SIZE0;
int  hint_size = HINT_SIZE0;
GC   auxgc;
Font fnt;

static char b0[] = "Exit";
static char b4[] = "Export";
static char b1a[] = "Choose";
static char b1b[] = "Edit hints";
static char b1c[] = "Play";
static char b2[] = "Permute";
static char b3[] = "Toss up";

static char iw1[] = "# hints";
static char swh[] = "*";
static char ErrMsgCannotExport[] = "Error: Cannot write the file.";

char *InfoMsg[5] =
  { "Sudoku HD (16 x 16) generator",
    "",
    "This program is a part of the BSTools package.",
    "(C) Copyright by Przemyslaw Kiciak, 2019",
    NULL };

xge_widget     *menu, *menu0wdg, *menu1wdg, *menu2wdg, *suwidget;
xge_int_widget nhintsw;
int            nhints, score;
char           unique = false;
char           show_eprompts = false;

int            nextrahints, nerrors;
char           strscore[32], strextrah[32], strerrors[32];

xge_filemenu   export_menu;
char           fullpath[XGE_MAX_PATH_LENGTH+1],
               shortpath[XGE_SHORT_PATH_LENGTH+1],
               prevpath[XGE_MAX_PATH_LENGTH+1],
               filter[XGE_MAX_FILENAME_LENGTH+1] = "*.tex",
               filename[XGE_MAX_FILENAME_LENGTH+1];
const char     file_ext[] = ".tex";


/* ////////////////////////////////////////////////////////////////////////// */
void r_InitPermut ( int l, int *i, char p[4] )
{
  int k, m;

  if ( l == 3 ) {
    memcpy ( permut[*i], p, 4*sizeof(char) );
    (*i) ++;
  }
  else {
    r_InitPermut ( l+1, i, p );
    for ( k = l+1; k < 4; k++ ) {
      m = p[l];  p[l] = p[k];  p[k] = m;
      r_InitPermut ( l+1, i, p );
      m = p[l];  p[l] = p[k];  p[k] = m;
    }
  }
} /*r_InitPermut*/

void InitPermut ( void )
{
  char p[4];
  int  i;

  for ( i = 0; i < 4; i++ )
    p[i] = i;
  i = 0;
  r_InitPermut ( 0, &i, p );
} /*InitPermut*/

void GenInitialSu ( char su[16][16] )
{
  int i, j, k;

  for ( i = 0; i < 16; i++ )
    su[0][i] = i;
  for ( j = 1; j < 4; j++ )
    for ( i = 0; i < 16; i++ )
      su[j][i] = (su[j-1][i] + 4) % 16;
  for ( k = 4; k < 16; k += 4 )
    for ( j = k; j < k+4; j++ )
      for ( i = 0; i < 16; i++ )
        su[j][i] = (su[j-4][i] + 1) % 16;
} /*GenInitialSu*/

void GenSu ( int param[10], char su[16][16] )
{
  int i, j, k;
  char s[16][16];
  char *p;

  GenInitialSu ( su );
  memcpy ( s, su, 256*sizeof(char) );
  p = permut[(int)param[0]];
  for ( i = 0; i < 4; i++ )
    for ( j = 0; j < 4; j++ )
      for ( k = 0; k < 16; k++ )
        su[4*p[i]+j][k] = s[4*i+j][k];
  for ( i = 0; i < 4; i++ ) {
    memcpy ( s, su, 256*sizeof(char) );
    p = permut[(int)param[i+1]];
    for ( j = 0; j < 4; j++ )
      for ( k = 0; k < 16; k++ )
        su[4*i+p[j]][k] = s[4*i+j][k];
  }
  memcpy ( s, su, 256*sizeof(char) );
  p = permut[(int)param[5]];
  for ( i = 0; i < 4; i++ )
    for ( j = 0; j < 4; j++ )
      for ( k = 0; k < 16; k++ )
        su[k][4*p[i]+j] = s[k][4*i+j];
  for ( i = 0; i < 4; i++ ) {
    memcpy ( s, su, 256*sizeof(char) );
    p = permut[(int)param[i+6]];
    for ( j = 0; j < 4; j++ )
      for ( k = 0; k < 16; k++ )
        su[k][4*i+p[j]] = s[k][4*i+j];
  }
  for ( i = 0; i < 16; i++ )
    for ( j = 0; j < 16; j++ )
      su[i][j] = hpermut[(int)su[i][j]];
} /*GenSu*/

void PrintSu ( char su[16][16] )
{
  int i, j;

  printf ( "\n" );
  for ( i = 0; i < 16; i++ ) {
    for ( j = 0; j < 16; j++ )
      printf ( "%c ", hd[(int)su[i][j]] );
    printf ( "\n" );
  }
  printf ( "\n" );
} /*PrintSu*/

/* ////////////////////////////////////////////////////////////////////////// */
void GenHPermut ( int param[10], char su[16][16] )
{
  struct tms   buffer;
  unsigned int tmp, rnd;
  int          i, sw;
  char         c;

  tmp = (unsigned int)times ( &buffer ) | 0x01;
  for ( i = 0; i < 10; i++ )
    tmp *= 16807;
  for ( i = 0; i < 16; i++ )
    hpermut[i] = i;
  for ( i = 0; i < 15; i++ ) {
    tmp *= 16807;
    rnd = tmp ^ (tmp >> 4) ^ (tmp << 15);
    sw = i + rnd % (16-i);
    c = hpermut[i];  hpermut[i] = hpermut[sw];  hpermut[sw] = c;
  }
/*
  printf ( "GenHPermut: " );
  for ( i = 0; i < 16; i++ )
    printf ( "%2d,", hpermut[i] );
  printf ( "\n" );
*/
  GenSu ( param, su );
} /*GenHPermut*/

void TossUp ( int param[10], char su[16][16] )
{
  struct tms   buffer;
  unsigned int tmp, rnd;
  int          i;

  tmp = (unsigned int)times ( &buffer ) | 0x01;
  for ( i = 0; i < 10; i++ )
    tmp *= 16807;
  for ( i = 0; i < 10; i++ ) {
    tmp *= 16807;
    rnd = tmp ^ (tmp >> 4) ^ (tmp << 15);
    param[i] = rnd % 24;
  }
  GenSu ( param, su );
} /*TosUp*/

/* ////////////////////////////////////////////////////////////////////////// */
void InitHints ( char sustatus[16][16] )
{
  memset ( sustatus, SU_NONE, 256*sizeof(char) );
} /*InitHints*/

boolean VerifySol ( char su[16][16], unsigned short prompts[16][16],
                    char tsol[16][16] )
{
  int i, j;

  for ( i = 0; i < 16; i++ )
    for ( j = 0; j < 16; j++ ) {
      if ( su[i][j] != tsol[i][j] ) {
        if ( !(prompts[i][j] & 0x0001 << su[i][j]) ) {
          printf ( "\n(%d,%d): %c, 0x%04x, %c(%d)\n",
                   i, j, hd[(int)su[i][j]], prompts[i][j],
                   tsol[i][j] >= 0 ? hd[(int)tsol[i][j]] : -1, tsol[i][j] );
          return false;
        }
      }
    }
  return true;
} /*VerifySol*/

void UpdatePrompts ( char su[16][16], char sustatus[16][16],
                     unsigned short prompts[16][16] )
{
  char tsol[16][16];

  TestHints ( su, sustatus, prompts, tsol, &unique );
  VerifySol ( su, prompts, tsol );
} /*UpdatePrompts*/

/* ////////////////////////////////////////////////////////////////////////// */
void InitPlay ( void )
{
  int i, j;

  for ( i = 0; i < 16; i++ )
    for ( j = 0; j < 16; j++ )
      if ( sustatus[i][j] == SU_HINT )
        suuser[i][j] = su[i][j];
      else
        { suuser[i][j] = -1;  sustatus[i][j] = SU_NONE; }
  score = nextrahints = nerrors = 0;
  sprintf ( strscore, "score     : %d", score );
  sprintf ( strextrah, "extra help: %d", nextrahints );
  sprintf ( strerrors, "errors    : %d", nerrors );
} /*InitPlay*/

void SuPlayMClick ( xge_widget *er, int key, short x, short y )
{
  int i, j;

  if ( key & xgemouse_RBUTTON_DOWN && key & xgemouse_RBUTTON_CHANGE ) {
    i = (y-er->y-CELL_Y)/cell_size;
    j = (x-er->x-CELL_X)/cell_size;
    if ( i >= 0 && i < 16 && j >= 0 && j < 16 )
      if ( sustatus[i][j] != SU_HINT ) {
        if ( sustatus[i][j] == SU_USER )
          sprintf ( strscore, "score     : %d", --score );
        suuser[i][j] = -1;
        sustatus[i][j] = SU_NONE;
        xge_Redraw ();
      }
  }
} /*SuPlayMClick*/

boolean IsIncorrect ( int i, int j, char dg )
{
  int k, l;

  for ( k = 0; k < 16; k++ )
    if ( k != j && suuser[i][k] == dg )
      return true;
  for ( k = 0; k < 16; k++ )
    if ( k != i && suuser[k][j] == dg )
      return true;
  for ( k = (i & 0xFC); k <= (i | 0x03); k++ )
    for ( l = (j & 0xFC); l <= (j | 0x03); l++ )
      if ( k != i && l != j && suuser[k][l] == dg )
        return true;
  return false;
} /*IsIncorrect*/

void SuPlayKey ( xge_widget *er, int key, short x, short y )
{
  int  i, j;
  char dg;

  if ( isxdigit ( key ) ) {
    if ( isdigit ( key ) )
      dg = key - '0';
    else
      dg = islower ( key ) ? key - 'a' + 10 : key - 'A' + 10;
    i = (y-er->y-CELL_Y)/cell_size;
    j = (x-er->x-CELL_X)/cell_size;
    if ( i >= 0 && i < 16 && j >= 0 && j < 16 ) {
      suuser[i][j] = dg;
      if ( IsIncorrect ( i, j, dg ) ) {
        sustatus[i][j] = SU_ERROR;
        nerrors ++;
        sprintf ( strerrors, "errors    : %d", nerrors );
      }
      else {
        sustatus[i][j] = SU_USER;
        score ++;
        sprintf ( strscore, "score     : %d", score );
      }
      xge_Redraw ();
    }
  }
} /*SuPlayKey*/

void CheckUserErrors ( char su[16][16], char suuser[16][16],
                       char sustatus[16][16] )
{
  int i, j;

  for ( i = 0; i < 16; i++ )
    for ( j = 0; j < 16; j++ )
      if ( sustatus[i][j] == SU_USER && suuser[i][j] != su[i][j] ) {
        sustatus[i][j] = SU_ERROR;
        nerrors ++;
        sprintf ( strerrors, "errors    : %d", nerrors );
      }
} /*CheckUserErrors*/

void SuPlaySpecialKey ( xge_widget *er, int key, short x, short y )
{
  int i, j;

  switch ( key ) {
case XK_F1:
    i = (y-er->y-CELL_Y)/cell_size;
    j = (x-er->x-CELL_X)/cell_size;
    if ( i >= 0 && i < 16 && j >= 0 && j < 16 )
      if ( sustatus[i][j] == SU_NONE || sustatus[i][j] == SU_ERROR ||
           sustatus[i][j] == SU_EPROMPT ) {
        sustatus[i][j] = SU_EHINT;
        suuser[i][j] = su[i][j];
        goto inc_nextra;
      }
    break;

case XK_F2:
    if ( (show_eprompts = ExtraPrompts ( suuser, sustatus )) ) {
inc_nextra:
      nextrahints ++;
      sprintf ( strextrah, "extra help: %d", nextrahints );
      xge_Redraw ();
    }
    break;

case XK_F3:
    CheckUserErrors ( su, suuser, sustatus );
    xge_Redraw ();
    break;

default:
    break;
  }
} /*SuPlaySpecialKey*/

/* ////////////////////////////////////////////////////////////////////////// */
void InitSUObject ( void )
{
  InitPermut ();
  GenSu ( par, su );
  InitHints ( sustatus );
  TossUpHints ( su, sustatus, hintpermut, hpinv, &nhints );
  UpdatePrompts ( su, sustatus, suprompts );
/*  PrintSu ( su );*/
} /*InitSUObject*/

xge_int_widget suparwdg[10];

void auxSetForeground ( unsigned int colour )
{
  if ( cell_size == CELL_SIZE0 )
    xgeSetForeground ( colour );
  else
    XSetForeground ( xgedisplay, auxgc, colour);
} /*auxSetForeground*/

void auxDrawString ( char *str, int x, int y )
{
  if ( cell_size == CELL_SIZE0 )
    xgeDrawString ( str, x, y );
  else
    XDrawString ( xgedisplay, xgepixmap, auxgc, x, y+8, str, strlen ( str ) );
} /*auxDrawString*/

void DisplaySu ( xge_widget *er )
{
  int i, j;
  char str[2];

  auxSetForeground ( xgec_White );
  str[1] = 0;
  for ( i = 0; i < 16; i++ )
    for ( j = 0; j < 16; j++ ) {
      str[0] = hd[(int)su[i][j]];
      auxDrawString ( str, er->x+CELL_X+10+cell_size*j,
                      er->y+CELL_Y+16+cell_size*i );
    }
} /*DisplaySu*/

void DisplaySuElements ( xge_widget *er, char s[16][16],
                         char status, unsigned int colour )
{
  int  i, j;
  char str[2];

  auxSetForeground ( colour );
  str[1] = 0;
  for ( i = 0; i < 16; i++ )
    for ( j = 0; j < 16; j++ )
      if ( sustatus[i][j] == status ) {
        str[0] = hd[(int)s[i][j]];
        auxDrawString ( str, er->x+CELL_X+10+cell_size*j,
                        er->y+CELL_Y+16+cell_size*i );
      }
} /*DisplaySuElements*/

void DisplaySuPrompts ( xge_widget *er )
{
  int            i, j, k, l;
  int            x, y;
  unsigned short mask;

  xgeSetForeground ( xgec_Grey4 );
  for ( i = 0; i < 16; i++ )
    for ( j = 0; j < 16;  j++ )
      if ( sustatus[i][j] != SU_HINT ) {
        y = er->y+CELL_Y+2+i*cell_size;
        x = er->x+CELL_X+2+j*cell_size;
        for ( k = 0, mask = 0x0001;  k < 4;  k++ )
          for ( l = 0;  l < 4;  l++, mask <<= 1 )
            if ( suprompts[i][j] & mask )
              xgeFillRectangle ( hint_size-1, hint_size-1,
                                 x+l*hint_size, y+k*hint_size );
      }
} /*DisplaySuPrompts*/

void DisplaySuEPrompts ( xge_widget *er )
{
  int i, j;

  xgeSetForeground ( xgec_Blue2 );
  for ( i = 0; i < 16; i++ )
    for ( j = 0; j < 16; j++ )
      if ( sustatus[i][j] == SU_EPROMPT )
        xgeFillRectangle ( cell_size-4, cell_size-4,
                           er->x+CELL_X+2+j*cell_size, er->y+CELL_Y+2+i*cell_size );
} /*DisplaySuEPrompts*/

void SuWinRedraw ( xge_widget *er, boolean onscreen )
{
  int i;

  if ( er->state == SU_HINTS && unique )
    xgeSetForeground ( xgec_Blue5 );
  else
    xgeSetForeground ( xgec_Blue4 );
  xgeFillRectangle ( er->w-2, er->h-2, er->x+1, er->y+1 );
  xgeSetForeground ( xgec_CornflowerBlue );
  xgeDrawRectangle ( er->w-1, er->h-1, er->x, er->y );
  xgeSetForeground ( xgec_Grey3 );
  for ( i = 0; i <= 16; i++ )
    xgeDrawLine ( er->x+CELL_X+cell_size*i, er->y+CELL_Y,
                  er->x+CELL_X+cell_size*i, er->y+CELL_Y+16*cell_size );
  for ( i = 0; i <= 16; i++ )
    xgeDrawLine ( er->x+CELL_X, er->y+CELL_Y+cell_size*i,
                  er->x+CELL_X+16*cell_size, er->y+CELL_Y+cell_size*i );
  xgeSetForeground ( xgec_White );
  for ( i = 0; i <= 16; i += 4 )
    xgeDrawLine ( er->x+CELL_X+cell_size*i, er->y+CELL_Y,
                  er->x+CELL_X+cell_size*i, er->y+CELL_Y+16*cell_size );
  for ( i = 0; i <= 16; i += 4 )
    xgeDrawLine ( er->x+CELL_X, er->y+CELL_Y+cell_size*i,
                  er->x+CELL_X+16*cell_size, er->y+CELL_Y+cell_size*i );
  switch ( er->state ) {
case SU_CHOOSE:
    DisplaySu ( er );
    break;
case SU_HINTS:
    DisplaySuElements ( er, su, SU_HINTS, xgec_White );
    DisplaySuPrompts ( er );
    break;
case SU_PLAY:
    DisplaySuElements ( er, suuser, SU_HINTS, xgec_White );
    DisplaySuElements ( er, suuser, SU_EHINT, xgec_GreenYellow );
    DisplaySuElements ( er, suuser, SU_USER, xgec_Grey3 );
    DisplaySuElements ( er, suuser, SU_ERROR, xgec_Red );
    DisplaySuEPrompts ( er );
    break;
default:
    break;
  }
  if ( onscreen )
    xgeCopyRectOnScreen ( er->w, er->h, er->x, er->y );
} /*SuWinRedraw*/

boolean ToggleHint ( int i, int j, boolean on )
{
  int k, l, m;

  k = 16*i+j;
  if ( on ) {
    if ( sustatus[i][j] == SU_HINT )
      return false;
    sustatus[i][j] = SU_HINT;
    l = hpinv[k];
    m = hintpermut[nhints];  hintpermut[nhints] = hintpermut[l];
    hintpermut[l] = m;
    hpinv[hintpermut[nhints]] = nhints;  hpinv[m] = l;
    nhints ++;
  }
  else {
    if ( sustatus[i][j] != SU_HINT )
      return false;
    sustatus[i][j] = SU_NONE;
    nhints --;
    l = hpinv[k];
    m = hintpermut[nhints];  hintpermut[nhints] = hintpermut[l];
    hintpermut[l] = m;
    hpinv[hintpermut[nhints]] = nhints;  hpinv[m] = l;
  }
  UpdatePrompts ( su, sustatus, suprompts );
  return true;
} /*ToggleHint*/

void SuWinToggleHint ( xge_widget *er, int msg, int key, short x, short y )
{
  int i, j;

  i = (y-er->y-CELL_Y)/cell_size;
  j = (x-er->x-CELL_X)/cell_size;
  if ( i >= 0 && i < 16 && j >= 0 && j < 16 ) {
    if ( key & xgemouse_LBUTTON_DOWN && key & xgemouse_LBUTTON_CHANGE ) {
      if ( !ToggleHint ( i, j, true ) )
        return;
    }
    else if ( key & xgemouse_RBUTTON_DOWN && key & xgemouse_RBUTTON_CHANGE ) {
      if ( !ToggleHint ( i, j, false ) )
        return;
    }
    else return;
    xge_Redraw ();
  }
} /*SuWinToggleHint*/

boolean SuWinMsg ( xge_widget *er, int msg, int key, short x, short y )
{
  if ( msg == xgemsg_RESIZE ) {
    er->w = 16*cell_size+2*CELL_X;
    er->h = 16*cell_size+2*CELL_Y;
  }
  else {
    switch ( er->state ) {
case SU_CHOOSE:
      break;
case SU_HINTS:
      switch ( msg ) {
  case xgemsg_MCLICK:
        SuWinToggleHint ( er, msg, key, x, y );
        break;
  default:
        break;
      }
      break;
case SU_PLAY:
      switch ( msg ) {
  case xgemsg_MCLICK:
        SuPlayMClick ( er, key, x, y );
        break;
  case xgemsg_KEY:
        SuPlayKey ( er, key, x, y );
        break;
  case xgemsg_SPECIAL_KEY:
        SuPlaySpecialKey ( er, key, x, y );
        break;
  default:
        break;
      }
      break;
default:
      break;
    }
  }
/*  printf ( "%d, %d, %d, %d, %d\n", er->id, msg, key, x, y );*/
  return true;
} /*SuWinMsg*/

/* ////////////////////////////////////////////////////////////////////////// */
char LoadFontFromPattern ( const char *pattern )
{
  char **fnames;
  int  fcnt;

  fnames = XListFonts ( xgedisplay, pattern, 1, &fcnt );
  if ( fcnt )
    fnt = XLoadFont ( xgedisplay, fnames[0] );
  XFreeFontNames ( fnames );
  return fcnt > 0;
} /*LoadFontFromPattern*/

void LoadBigFont ( void )
{
  auxgc = XCreateGC ( xgedisplay, xgewindow, 0, 0 );
  if ( LoadFontFromPattern ( "*courier-medium-r*180*" ) ) ;
  else if ( LoadFontFromPattern ( "*utopia-regular-r*180*" ) ) ;
  else if ( LoadFontFromPattern ( "*times-medium-r*180*" ) ) ;
  else if ( LoadFontFromPattern ( "*helvetica-medium-r*180*" ) ) ;
  else if ( LoadFontFromPattern ( "*lucida-medium-r*180*" ) ) ;
  else if ( LoadFontFromPattern ( "*lucidatypewriter-medium-r*180*" ) ) ;
  else if ( LoadFontFromPattern ( "*fixed-medium-r*130*" ) ) ;
  else if ( !LoadFontFromPattern ( "fixed" ) ) {
    printf ( "Cannot load any suitable font!\n" );
    exit ( 1 );
  }
  XSetFont ( xgedisplay, auxgc, fnt );
} /*LoadBigFont*/

void init_edwin ( void )
{
  xge_widget *rp;
  int        i, j, k;

  LoadBigFont ();
  XSetForeground ( xgedisplay, xgegc, xgec_White );
  InitSUObject ();
  rp = xge_NewButton ( 0, NULL, BTN_EXIT, 70, 18, 0, 0, b0 );
  rp = xge_NewButton ( 0, rp, BTN_EXPORT, 70, 18, 0, 20, b4 );
  rp = xge_NewButton ( 0, rp, BTN_HINTS, 70, 18, 0, 40, b1b );
  rp = xge_NewButton ( 0, rp, BTN_TOSS, 58, 18, 0, 68, b3 );
  for ( i = k = 0;  i < 2;  i++ )
    for ( j = 0;  j < 5;  j++, k++ )
      rp = xge_NewIntWidget ( 0, rp, INTW_CHOOSE+k, 18, 18, 20*j, 88+20*i, 0, 23,
                              &suparwdg[k], NULL, &par[k] );
  rp = xge_NewButton ( 0, rp, BTN_PERMUTE, 58, 18, 0, 128, b2 );
  menu = xge_NewMenu ( 0, NULL, 1, MENU_WIDTH, xge_HEIGHT, 0, 0, menu0wdg = rp );

  rp = xge_NewButton ( 0, NULL, BTN_EXIT, 70, 18, 0, 0, b0 );
  rp = xge_NewButton ( 0, rp, BTN_EXPORT, 70, 18, 0, 20, b4 );
  rp = xge_NewButton ( 0, rp, BTN_PLAY, 70, 18, 0, 40, b1c );
  rp = xge_NewButton ( 0, rp, BTN_TOSSH, 70, 18, 0, 68, b3 );
  rp = xge_NewIntWidget ( 0, rp, INTW_NHINTS, 70, 18, 0, 88,
                          0, 256, &nhintsw, iw1, &nhints );
  menu1wdg = xge_NewSwitch ( 0, rp, SW_STAR_HINTS, 24, 16, 72, 89,
                             swh, &star_hints );

  rp = xge_NewButton ( 0, NULL, BTN_EXIT, 70, 18, 0, 0, b0 );
  rp = xge_NewButton ( 0, rp, BTN_EXPORT, 70, 18, 0, 20, b4 );
  rp = xge_NewButton ( 0, rp, BTN_CHOOSE, 70, 18, 0, 40, b1a );
  rp = xge_NewTextWidget ( 0, rp, 0, 120, 18, 0, 68, strscore );
  rp = xge_NewTextWidget ( 0, rp, 0, 120, 18, 0, 88, strextrah );
  rp = xge_NewTextWidget ( 0, rp, 0, 120, 18, 0, 108, strerrors );
  menu2wdg = rp;

  xge_InitSaveFileMenu ( 0, MENU_EXPORTPOPUP, 400, 232, 20, 40,
                         &export_menu, NULL, fullpath, shortpath, 
                         prevpath, filter, filename );

  suwidget = xge_NewWidget ( 0, menu, 0, xge_WIDTH-MENU_WIDTH, xge_HEIGHT,
                             MENU_WIDTH, 0, NULL, NULL, SuWinMsg, SuWinRedraw );
  suwidget->state = SU_CHOOSE;
  xge_SetWinEdRect ( suwidget );

  xge_Redraw ();
  xge_DisplayInfoMessage ( InfoMsg, -1 );
} /*init_edwin*/

void destroy_edwin ( void )
{
  XFreeGC ( xgedisplay, auxgc );
  XUnloadFont ( xgedisplay, fnt );
} /*destroy_edwin*/

int CallBack ( xge_widget *er, int msg, int key, short x, short y )
{
  int i, j;

  if ( er ) {
    switch ( msg ) {
case xgemsg_BUTTON_COMMAND:
      switch ( er->id ) {
  case BTN_EXIT:
        xge_done = 1;
        break;
  case BTN_HINTS:
        TossUpHints ( su, sustatus, hintpermut, hpinv, &nhints );
        UpdatePrompts ( su, sustatus, suprompts );
        xge_SetMenuWidgets ( menu, menu1wdg, false );
        suwidget->state = SU_HINTS;
        xge_Redraw ();
        return 1;
  case BTN_CHOOSE:
        xge_SetMenuWidgets ( menu, menu0wdg, false );
        suwidget->state = SU_CHOOSE;
        xge_Redraw ();
        return 1;
  case BTN_PLAY:
        xge_SetMenuWidgets ( menu, menu2wdg, false );
        suwidget->state = SU_PLAY;
        InitPlay ();
        xge_Redraw ();
        return 1;
  case BTN_PERMUTE:
        GenHPermut ( par, su );
        xge_Redraw ();
        return 1;
  case BTN_TOSS:
        TossUp ( par, su );
        xge_Redraw ();
        return 1;
  case BTN_TOSSH:
        TossUpHints ( su, sustatus, hintpermut, hpinv, &nhints );
        UpdatePrompts ( su, sustatus, suprompts );
        xge_Redraw ();
        return 1;
  case BTN_EXPORT:
        xge_AddSaveFilePopup ( &export_menu );
        return 1;
  default:
        break;
      }
      break;

case xgemsg_SWITCH_COMMAND:
      switch ( er->id ) {
  case SW_STAR_HINTS:
        UpdatePrompts ( su, sustatus, suprompts );
        xge_Redraw ();
        return 1;
  default:
        break;
      }
      break;
    
case xgemsg_INT_WIDGET_COMMAND:
      if ( er->id >= INTW_CHOOSE && er->id < INTW_CHOOSE+10 ) {
        par[er->id-2] = key;
        GenSu ( par, su );
        xge_Redraw ();
        return 1;
      }
      else if ( er->id == INTW_NHINTS ) {
        if ( key > nhints ) {
          for ( i = nhints; i < key; i++ ) {
            j = hintpermut[i];
            sustatus[j/16][j%16] = SU_HINT;
          }
        }
        else if ( key >= 0 && key < nhints ) {
          for ( i = nhints-1; i >= key; i-- ) {
            j = hintpermut[i];
            sustatus[j/16][j%16] = SU_NONE;
          }
        }
        else break;
        nhints = key;
        UpdatePrompts ( su, sustatus, suprompts );
        xge_Redraw ();
        return 1;
      }
      break;

case xgemsg_FILEMENU_SAVEFILE:
      switch ( er->id ) {
  case MENU_EXPORTPOPUP:
        xge_RemovePopup ( true );
        if ( !ExportSu ( filename, su, sustatus ) )
          xge_DisplayErrorMessage ( ErrMsgCannotExport, -1 );
        return 1;
      }
      break;

default:
      break;
    }
  }
  else {
/*    printf ( "%d, %d, %d, %d\n", msg, key, x, y );*/
    switch ( msg ) {
case xgemsg_RESIZE:
      menu->msgproc ( menu, xgemsg_RESIZE, 0, MENU_WIDTH, y );
      suwidget->msgproc ( suwidget, xgemsg_RESIZE, 0, x-MENU_WIDTH, y );
      xge_Redraw ();
      break;

case xgemsg_KEY:
      if ( key == 'm' || key == 'M' ) {
        if ( cell_size == CELL_SIZE0 ) {
          cell_size = CELL_SIZE1;
          hint_size = HINT_SIZE1;
        }
        else {
          cell_size = CELL_SIZE0;
          hint_size = HINT_SIZE0;
        }
        xgeResizeWindow ( MENU_WIDTH+16*cell_size+2*CELL_X,
                          16*cell_size+2*CELL_Y );
      }
      break;

default:
      break;
    }
  }
  return 0;
} /*CallBack*/

/* ////////////////////////////////////////////////////////////////////////// */
int main ( int argc, char **argv )
{
  if ( !xge_Init ( argc, argv, CallBack, NULL ) )
    exit ( 1 );
  init_edwin ();
  xge_MessageLoop ();
  destroy_edwin ();
  xge_Cleanup ();
  exit ( 0 );
} /*main*/

