
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2015                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#ifndef AVLTREE_H
#define AVLTREE_H

#ifdef __cplusplus
extern "C" {
#endif

#define AVLERROR_NONE       0
#define AVLERROR_ALLOC      1
#define AVLERROR_DOUBLE_KEY 2
#define AVLERROR_DELETE     3
#define AVLERROR_COMPARE    4
 
typedef struct AVLtnode {
    struct AVLtnode *left, *right;
    char eqi;
  } AVLtnode;

typedef struct AVLtree {
    AVLtnode *root;
    AVLtnode *last;
    boolean  no_double;
    char     error_code;
    char     (*compkeys)( AVLtnode *node, void *data );
    AVLtnode *(*allocnode) ( void *data );
    void     (*freenode) ( AVLtnode *node );
  } AVLtree;

/* ///////////////////////////////////////////////////////////////// */
/* An application must provide three procedures:                     */
/* the compkeys procedure must return -1 if the key in the tree node */
/* pointed by the first parameter is less than the key in the data   */
/* pointed by the second parameter, 0 if the keys are equal and +1   */
/* if it is greater                                                  */
/* the allocnode procedure must allocate the memory for a new node,  */
/* then copy the key and other data to this node and return its      */
/* address, which may be NULL if the allocation failed.              */
/* the freenode procedure must deallocate the node                   */
void AVLInit ( AVLtree *tree, boolean no_double,
               char (*compkeys)(AVLtnode*,void*),
               AVLtnode *(*allocnode)(void*),
               void (*freenode)(AVLtnode*) );

boolean AVLInsert ( AVLtree *tree, void *data );

boolean AVLDelete ( AVLtree *tree, void *data );

AVLtnode *AVLFind ( AVLtree *tree, void *data );

void AVLTraverse ( AVLtree *tree, boolean descending, void *usrdata,
                   void (*process)(void*,AVLtnode*) );

/* deallocates only nodes */
void AVLDestroy ( AVLtree *tree );

#ifdef __cplusplus
}
#endif

#endif /*AVLTREE_H*/

