
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2010, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#ifndef XGLEDIT_H
#define XGLEDIT_H

#ifndef XGLE_OPENGL_OLD
#define XGLE_OPENGL_NEW
#endif

#ifndef _LIBC_LIMITS_H_
#include <limits.h>
#endif

#ifndef __gl_h_
#include <GL/gl.h>
#include <GL/glxext.h>
#endif
#ifndef GLX_H
#include <GL/glx.h>
#endif

#ifndef PKVARIA_H
#include "pkvaria.h"
#endif
#ifndef PKNUM_H
#include "pknum.h"
#endif
#ifndef PKGEOM_H
#include "pkgeom.h"
#endif
#ifndef CAMERA_H
#include "camera.h"
#endif
#ifndef XGEDIT_H
#include "xgedit.h"
#endif
#ifndef XGLERGB_H
#include "xgedit/xglergb.h"
#endif

#define XGLE_DEPTH_BUFFER    0x01
#define XGLE_STENCIL_BUFFER  0x02
#define XGLE_ACCUM_BUFFER    0x04  /* not available in the new OpenGL */

#ifdef __cplusplus
extern "C" {
#endif

#define xgleCopyGLRect(w,h,x,y) \
  XCopyArea(xgedisplay,xglepixmap,xgepixmap,xgegc,0,xge_MAX_HEIGHT-h,w,h,x,y)
#define xgleClearColor3fv(rgb) \
  glClearColor(rgb[0],rgb[1],rgb[2],1.0)

extern Pixmap     xglepixmap;
extern XID        _xglepixmap;
extern GLXContext xglecontext;

/* OpenGL context version */
extern byte       _xgle_majorgl, _xgle_minorgl;
extern boolean    _xgle_ctx_direct;
/* information about available GL resources, left by xgle_Init */
extern boolean    _xgle_depth, _xgle_accum, _xgle_stencil;

extern GLfloat xgle_palette[XGLE_PALETTE_LENGTH][3];

extern const char *GLProcName[];

/* ////////////////////////////////////////////////////////////////////////// */
boolean xgle_IsExtensionSupported ( const char *extList, const char *extension );
proc_addr _xgle_GetProcAddr ( const char *name );
void _xgle_GetGLProcedures ( void );

boolean xgle_Init ( int argc, char *argv[],
                    int (*callback)(xge_widget*,int,int,short,short),
                    char *title,
                    int majorgl, int minorgl,
                    unsigned int required, unsigned int nice );
void xgle_Cleanup ( void );

void xgle_ExitOnGLError ( const char *msg );
void xgle_IgnoreGLErrors ( boolean output );

/* ////////////////////////////////////////////////////////////////////////// */
void xgle_DrawGeomWinBackground ( xge_widget *er, GLbitfield mask );

void xgle_Trans3toGLf ( trans3f *tr, GLfloat a[16] );
void xgle_Trans3toGLd ( trans3d *tr, GLdouble a[16] );

void xgle_Frustumf ( float left, float right, float bottom, float top,
                     float near, float far, GLfloat m[16], GLfloat im[16] );
void xgle_Frustumd ( double left, double right, double bottom, double top,
                     double near, double far, GLdouble m[16], GLdouble im[16] );

void xgle_Orthof ( float left, float right, float bottom, float top,
                   float near, float far, float m[16], float im[16] );
void xgle_Orthod ( double left, double right, double bottom, double top,
                   double near, double far, double m[16], double im[16] );

boolean xgle_SetGLCameraf ( CameraRecf *CPos, GLint viewport[4],
                            GLfloat view[16], GLfloat projection[16],
                            GLfloat eyepos[4] );
boolean xgle_SetGLCamerad ( CameraRecd *CPos, GLint viewport[4],
                            GLdouble view[16], GLdouble projection[16],
                            GLdouble eyepos[4] );

#ifdef XGLE_OPENGL_OLD
/* ////////////////////////////////////////////////////////////////////////// */
/* the procedures below use "old style" OpenGL commands, i.e. the immediate   */
/* mode, deprecated in OpenGL standards newer than 2.1.                       */
void xgl2e_SetIdentMapping ( xge_widget *er );

boolean xgl2e_SetGLCameraf ( CameraRecf *CPos );
boolean xgl2e_SetGLCamerad ( CameraRecd *CPos );

boolean xgl2e_SetGLaccCameraf ( CameraRecf *CPos,
                                float pixdx, float pixdy,
                                float eyedx, float eyedy, float focus );
boolean xgl2e_SetGLaccCamerad ( CameraRecd *CPos,
                                double pixdx, double pixdy,
                                double eyedx, double eyedy, double focus );

void xgl2e_MultMatrix3f ( trans3f *tr );
void xgl2e_MultMatrix3d ( trans3d *tr );

void xgl2eDrawPoint ( int x, int y );
void xgl2eDrawPoints ( int n, XPoint *p );
void xgl2eDrawLine ( int x0, int y0, int x1, int y1 );
void xgl2eDrawRectangle ( int w, int h, int x, int y );
void xgl2eDrawString ( char *string, int x, int y );



void xgl2e_2DwinfDrawCursorPos ( xge_2Dwinf *_2Dwin, short x, short y );
void xgl2e_2DwinfDrawAxes ( xge_2Dwinf *_2Dwin );
void xgl2e_2DwindDrawCursorPos ( xge_2Dwind *_2Dwin, short x, short y );
void xgl2e_2DwindDrawAxes ( xge_2Dwind *_2Dwin );

void xgl2e_3DwinfDrawCursorPos ( xge_3Dwinf *_3Dwin,
                                 int id, short x, short y );
void xgl2e_3DwindDrawCursorPos ( xge_3Dwind *_3Dwin,
                                 int id, short x, short y );

void xgl2e_T2KnotWinfDrawCursorPos ( xge_T2KnotWinf *T2win, short x, short y );
void xgl2e_T2KnotWindDrawCursorPos ( xge_T2KnotWind *T2win, short x, short y );
#endif

#ifdef XGLE_OPENGL_NEW
#ifndef XGLEOPENGL_H
#include "xgedit/xgleopengl.h"
#endif

/* OpenGL uniform block binding points - see xgleprivate.h */
#define XGLE_LAST_BINDPOINT 4

/* ////////////////////////////////////////////////////////////////////////// */
/* the procedures below are to be used in "new style" OpenGL applications.    */
GLuint xgle_CompileShaderStrings ( GLenum shader_type,
                                   int nsl, const GLchar *const*srclines );
GLuint xgle_CompileShaderFiles ( GLenum shader_type,
                                 int nfiles, const char *const*filenames );
GLuint xgle_LinkProgram ( int nsh, const GLuint *shaders );

/* these are convenience procedures implementing 2D graphics in a window */
/* using the new OpenGL. At least OpenGL 4.0 is required due to the GLSL */
/* version used by the shaders implementing the necessary programs.      */
boolean xgle_InitPrivateGLObjects ( void );
void xgle_SetupXGLCoordinates ( GLint w, GLint h );
void xgle_SetForeground ( unsigned int foreground, byte alpha );
void xgle_SetBackground ( unsigned int background, byte alpha );
void xgle_DrawLine ( GLfloat x0, GLfloat y0, GLfloat x1, GLfloat y1 );
void xgle_Choose12x6font ( void );
void xgle_Choose18x10font ( void );
void xgle_DrawString ( const char *string, int x, int y );

void xgle_2DwinfDrawCursorPos ( xge_2Dwinf *_2Dwin, short x, short y );
void xgle_2DwinfDrawAxes ( xge_2Dwinf *_2Dwin );

void xgle_2DwindDrawCursorPos ( xge_2Dwind *_2Dwin, short x, short y );
void xgle_2DwindDrawAxes ( xge_2Dwind *_2Dwin );

void xgle_3DwinfDrawCursorPos ( xge_3Dwinf *_3Dwin,
                                int id, short x, short y );
void xgle_3DwindDrawCursorPos ( xge_3Dwind *_3Dwin,
                                int id, short x, short y );

void xgle_T2KnotWinfDrawCursorPos ( xge_T2KnotWinf *T2win, short x, short y );
void xgle_T2KnotWindDrawCursorPos ( xge_T2KnotWind *T2win, short x, short y );

#endif

#ifdef __cplusplus
}
#endif

#endif /*XGLEDIT_H*/

