
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2014, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#ifndef PKSCANNER_H
#include "pkscanner.h"
#endif

#ifndef BSFPRIVATE_H
#define BSFPRIVATE_H

/* ////////////////////////////////////////////////////////////////////////// */
/* the subsequent keyword identifiers below must */
/* match the strings in the table in bsfile00r.c */
#define BSF_KEY_BCURVE             0  /* consecutive numbers of keywords */
#define BSF_KEY_BPATCH             1   /* sorted alphabetically, with uppercase */
#define BSF_KEY_BSCURVE            2   /* preceding all lowercase letters (ASCII) */
#define BSF_KEY_BSHOLE             3
#define BSF_KEY_BSMESH             4
#define BSF_KEY_BSPATCH            5
#define BSF_KEY_EULERANGLES        6
#define BSF_KEY_KLINKAGE           7
#define BSF_KEY_ART_PARAM          8
#define BSF_KEY_ART_PARAM_RANGE    9
#define BSF_KEY_BTRANS            10
#define BSF_KEY_CAMERA            11
#define BSF_KEY_CLOSED            12
#define BSF_KEY_COLOR             13   /* it can read both, AE and BE, */
#define BSF_KEY_COLOUR            14  /* but it will write in BE */
#define BSF_KEY_CPOINTS           15
#define BSF_KEY_CPOINTSMK         16
#define BSF_KEY_DEGREE            17
#define BSF_KEY_DEPTH             18
#define BSF_KEY_DIM               19
#define BSF_KEY_DOMAIN            20
#define BSF_KEY_ETRANS            21
#define BSF_KEY_FACETMK           22
#define BSF_KEY_FACETS            23
#define BSF_KEY_FILENAME          24
#define BSF_KEY_FRAME             25
#define BSF_KEY_FTRANS            26
#define BSF_KEY_HALFEDGEMK        27
#define BSF_KEY_HALFEDGES         28
#define BSF_KEY_IDENT             29
#define BSF_KEY_JOINT             30
#define BSF_KEY_KNOTS             31
#define BSF_KEY_KNOTS_U           32
#define BSF_KEY_KNOTS_V           33
#define BSF_KEY_LINK              34
#define BSF_KEY_NAME              35
#define BSF_KEY_NUM_ART_PAR       36
#define BSF_KEY_NUM_JOINTS        37
#define BSF_KEY_NUM_LINKS         38
#define BSF_KEY_NUM_OBJ_REF       39
#define BSF_KEY_NUM_OBJECTS       40
#define BSF_KEY_NUM_VERTICES      41
#define BSF_KEY_OBJECT            42
#define BSF_KEY_OBJECT_REF        43
#define BSF_KEY_PARALLEL          44
#define BSF_KEY_PERSPECTIVE       45
#define BSF_KEY_POINTS            46
#define BSF_KEY_POLYLINE          47
#define BSF_KEY_POSITION          48
#define BSF_KEY_RATIONAL          49
#define BSF_KEY_ROOT              50
#define BSF_KEY_ROTATION_V        51
#define BSF_KEY_ROTATION_X        52
#define BSF_KEY_ROTATION_Y        53
#define BSF_KEY_ROTATION_Z        54
#define BSF_KEY_SCALING_X         55
#define BSF_KEY_SCALING_XYZ       56
#define BSF_KEY_SCALING_Y         57
#define BSF_KEY_SCALING_Z         58
#define BSF_KEY_SIDES             59
#define BSF_KEY_SPHERICAL_PRODUCT 60
#define BSF_KEY_TRANSFORM         61
#define BSF_KEY_TRANSLATION_V     62
#define BSF_KEY_TRANSLATION_X     63
#define BSF_KEY_TRANSLATION_Y     64
#define BSF_KEY_TRANSLATION_Z     65
#define BSF_KEY_TRIMMED           66
#define BSF_KEY_UNIFORM           67
#define BSF_KEY_VERTICES          68
#define BSF_NKEYWORDS             69  /* number of the last keyword */

extern const char *bsf_keyword[BSF_NKEYWORDS];


/* ////////////////////////////////////////////////////////////////////////// */
boolean _bsf_OpenInputFile ( const char *filename, bsf_Reader *reader );

boolean bsf_ReadBezierCurve4d ( bsf_Reader *reader,
                                int maxdeg, int *deg, point4d *cpoints,
                                int *spdimen, boolean *rational,
                                char *name, int *ident );
boolean bsf_ReadBSplineCurve4d ( bsf_Reader *reader,
                                 int maxdeg, int maxlastknot, int maxncpoints,
                                 int *deg, int *lastknot, double *knots,
                                 boolean *closed, point4d *cpoints,
                                 int *spdimen, boolean *rational,
                                 char *name, int *ident );
boolean bsf_ReadBezierPatch4d ( bsf_Reader *reader,
                                int maxdeg, int *udeg, int *vdeg,
                                int *pitch, point4d *cpoints,
                                int *spdimen, boolean *rational,
                                char *name, int *ident );
boolean bsf_ReadBSplinePatch4d ( bsf_Reader *reader,
                                 int maxdeg, int maxlastknot, int maxncpoints,
                                 int *udeg, int *lastknotu, double *knotsu,
                                 int *vdeg, int *lastknotv, double *knotsv,
                                 boolean *closed_u, boolean *closed_v,
                                 int *pitch, point4d *cpoints,
                                 int *spdimen, boolean *rational,
                                 char *name, int *ident );
boolean bsf_ReadBSMesh4d ( bsf_Reader *reader,
                           int maxnv, int maxnhe, int maxnfac,
                           int *degree,
                           int *nv, BSMvertex *mv, int *mvhei, point4d *vc,
                           int *nhe, BSMhalfedge *mhe,
                           int *nfac, BSMfacet *mfac, int *mfhei,
                           int *spdimen, boolean *rational,
                           char *name, int *ident );
boolean bsf_ReadBSplineHoled ( bsf_Reader *reader,
                               int maxk, int *hole_k, double *knots,
                               point2d *domain_cp, point4d *hole_cp,
                               int *spdimen, boolean *rational,
                               char *name, int *ident );
boolean bsf_ReadKLinkage ( bsf_Reader *reader, pkkl_linkage *lkg,
                           int *objvert, char *name, int *ident );
boolean bsf_ReadPolyline4d ( bsf_Reader *reader,
                             int maxvert, int *nvert, point4d *vert,
                             int *spdimen, boolean *rational, boolean *closed,
                             char *name, int *ident );

boolean bsf_ReadTrimmedDomaind ( bsf_Reader *reader );
boolean bsf_ReadDependencies ( bsf_Reader *reader );

boolean bsf_ReadCamera ( bsf_Reader *reader, CameraRecd *Camera, int *ident );
boolean bsf_ReadColour ( bsf_Reader *reader, point3d *colour );

/* ////////////////////////////////////////////////////////////////////////// */
/* auxiliary, low-level reading procedures */
int bsf_GetNextSymbol ( pks_scanner *sc );

boolean bsf_ReadIntNumber ( pks_scanner *sc, int *number );
boolean bsf_ReadDoubleNumber ( pks_scanner *sc, double *number );
boolean bsf_ReadPointd ( pks_scanner *sc,
                         int maxcpdimen, double *point, int *cpdimen );
boolean bsf_ReadIdent ( pks_scanner *sc, int *ident );
int bsf_ReadPointsd ( pks_scanner *sc, int maxcpdimen, int maxnpoints,
                      double *points, int *cpdimen );
boolean bsf_ReadTrans3d ( pks_scanner *sc, trans3d *tr );
int bsf_ReadPointsMK ( pks_scanner *sc, int maxnpoints, unsigned int *mk );
int bsf_ReadHEdgeMK ( pks_scanner *sc, int maxnhe, unsigned int *hemk );
int bsf_ReadFacetMK ( pks_scanner *sc, int maxnfac, unsigned int *fmk );

boolean bsf_ReadSpaceDim ( pks_scanner *sc, int maxdim, int *spdimen );
boolean bsf_ReadCurveDegree ( pks_scanner *sc, int maxdeg, int *degree );
boolean bsf_ReadPatchDegree ( pks_scanner *sc, int maxdeg, int *udeg, int *vdeg );

boolean bsf_ReadKnotSequenced ( pks_scanner *sc, int maxlastknot,
                                int *lastknot, double *knots, boolean *closed );

boolean _bsf_ReadCPMark ( bsf_Reader *reader, int maxnpoints );
boolean _bsf_ReadHEdgeMark ( bsf_Reader *reader );
boolean _bsf_ReadFacetMark ( bsf_Reader *reader );
boolean _bsf_ReadColour ( bsf_Reader *reader );

/* ////////////////////////////////////////////////////////////////////////// */
/* output file formatting procedures and macros */

#define BSF_OUTPUT_LINE_LENGTH 64

void bsf_EndOutputLine ( bsf_Writer *writer );
void bsf_DivideOutputLine ( bsf_Writer *writer );
void bsf_WriteCurrentIndentation ( bsf_Writer *writer );

#define BSFeol bsf_EndOutputLine ( writer );
#define BSFdol bsf_DivideOutputLine ( writer );
#define BSFwci bsf_WriteCurrentIndentation ( writer );

#endif

