
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2005, 2015                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdio.h>  /* for testing */

#include "pkvaria.h"
#include "pkgeom.h"
#include "multibs.h"

#include "msgpool.h"

/* /////////////////////////////////////////// */
boolean mbs_FindPolynomialZerosf ( int degree, const float *coeff,
                                   int *nzeros, float *zeros, float eps )
{
  void          *sp;
  float        *a, *b, x;
  int           nz, stp;
  _mbs_BezPolyf ap;
  boolean       error, nonzero;

  sp = pkv_GetScratchMemTop ();

        /* push the interval [0,1] and the related polynomial coefficients */
  if ( !(a = pkv_GetScratchMemf ( degree+3 )) )
    goto failure;
  ap.deg = degree;
  memcpy ( &a[2], coeff, (degree+1)*sizeof(float) );
  a[0] = 0.0;
  a[1] = 1.0;
  stp = 1;

  nz = 0;
  do {
    stp --;
    if ( pkn_VarSignf ( degree+1, &a[2] ) ) {
      if ( pkn_OneSignChangef ( degree+1, &a[2], &nonzero ) ) {
          /* one sign change - try to solve */
        ap.coeff = &a[2];
        x = pkn_Illinoisf ( _mbs_EvalBezPolyf, (void*)&ap, 0.0, 1.0, eps, &error );
        if ( error )
          goto divide;
        zeros[nz++] = (1.0-x)*a[0] + x*a[1];
        if ( nz > degree )
          goto failure;
        pkv_FreeScratchMemf ( degree+3 );
        a -= degree+3;
      }
      else {  /* more than one sign change - bisect and push */
divide:
        if ( !(b = pkv_GetScratchMemf ( degree+3 )) )
          goto failure;
        stp += 2;
        b[0] = a[0];
        a[0] = b[1] = 0.5*(b[0]+a[1]);
        mbs_BisectBC1f ( degree, &a[2], &b[2] );
        a = b;
      }
    }
    else {    /* no sign change - reject the interval */
      pkv_FreeScratchMemf ( degree+3 );
      a -= degree+3;
    }
  } while ( stp != 0 );

  *nzeros = nz;
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*mbs_FindPolynomialZerosf*/

