
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2015                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <limits.h>
#include <string.h>
#include <math.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "multibs.h"

void mbs_multiRollClosedBSCurvesf ( int ncurves, int spdimen,
                                    int degree, int lastknot, float *knots,
                                    int pitch, float *ctlpoints,
                                    int intv,
                                    float *outknots, float *outctlpoints )
{
  int    clcK, i, j;
  float clcT;

  clcK = lastknot-2*degree;
  if ( knots )
    clcT = knots[lastknot-degree]-knots[degree];
  intv %= clcK;
  if ( !intv ) { /* just copy the knots and control points */
    if ( knots && outknots )
      memcpy ( outknots, knots, (lastknot+1)*sizeof(float) );
    pkv_Selectf ( ncurves, (lastknot-degree)*spdimen, pitch, pitch,
                  ctlpoints, outctlpoints );
  }
  else {
    if ( knots && outknots ) {
      for ( i = 0, j = intv; j <= lastknot; i++, j++ )
        outknots[i] = knots[j];
      for ( j = i+intv-clcK; i <= lastknot; i++, j++  )
        outknots[i] = knots[j]+clcT;
    }
    for ( i = 0, j = intv; j < lastknot-degree; i++, j++ )
      pkv_Selectf ( ncurves, spdimen, pitch, pitch,
                    &ctlpoints[j*spdimen], &outctlpoints[i*spdimen] );
    for ( j = i+intv-clcK; i < lastknot-degree; i++, j++ )
      pkv_Selectf ( ncurves, spdimen, pitch, pitch,
                    &ctlpoints[j*spdimen], &outctlpoints[i*spdimen] );
  }
} /*mbs_multiRollClosedBSCurvesf*/

