
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017, 2018                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "pklinkage.h"

/* ///////////////////////////////////////////////////////////////////////// */
pkkl_linkage *pkkl_NewLinkage ( int maxo, int maxl, int maxr,
                                int maxj, int maxp, double *artprange,
                                void *usrdata )
{
  pkkl_linkage *lkg;
  int          i;

  lkg = malloc ( sizeof(pkkl_linkage) );
  if ( lkg ) {
    memset ( lkg, 0, sizeof(pkkl_linkage) );
    lkg->maxobj    = maxo;
    lkg->maxlinks  = maxl;
    lkg->maxorefs  = maxr;
    lkg->maxnhj    = 2*maxj;
    lkg->maxartpar = maxp;
    lkg->obj  = malloc ( maxo*sizeof(pkkl_geom_obj) );
    lkg->link = malloc ( maxl*sizeof(pkkl_link) );
    lkg->oref = malloc ( maxr*sizeof(pkkl_obj_ref) );
    lkg->hj   = malloc ( 2*maxj*sizeof(pkkl_halfjoint) );
    lkg->artp = malloc ( 4*maxp*sizeof(double) );
    if ( !lkg->obj || !lkg->link || !lkg->oref || !lkg->hj || !lkg->artp )
      goto failure;
    lkg->prevartp = &lkg->artp[maxp];
    lkg->artprange = &lkg->prevartp[maxp];
    memset ( lkg->artp, 0, maxp*sizeof(double) );
    memset ( lkg->prevartp, 0x3f, maxp*sizeof(double) );
    if ( artprange )
      memcpy ( lkg->artprange, artprange, 2*maxp*sizeof(double) );
    else {
      for ( i = 0; i < maxp; i++ )
        { lkg->artprange[2*i] = 0.0;  lkg->artprange[2*i+1] = 1.0; }
    }
    lkg->current_root = 0;
    IdentTrans3d ( &lkg->current_root_tr );
    lkg->usrdata = usrdata;
  }
  return lkg;

failure:
  if ( lkg->obj )  free ( lkg->obj );
  if ( lkg->link ) free ( lkg->link );
  if ( lkg->oref ) free ( lkg->oref );
  if ( lkg->hj )   free ( lkg->hj );
  if ( lkg->artp ) free ( lkg->artp );
  free ( lkg );
  return NULL;
} /*pkkl_NewLinkage*/

void pkkl_DestroyLinkage ( pkkl_linkage *linkage )
{
  int i;

  if ( linkage->obj ) {
    for ( i = 0; i < linkage->nobj; i++ ) {
      if ( linkage->obj[i].destroy )
        linkage->obj[i].destroy ( &linkage->obj[i] );
      if ( linkage->obj[i].name )
        free ( linkage->obj[i].name );
      if ( linkage->obj[i].filename )
        free ( linkage->obj[i].filename );
    }
    free ( linkage->obj );
  }
  if ( linkage->oref ) {
    for ( i = 0; i < linkage->norefs; i++ )
      if ( linkage->oref[i].vn )
        free ( linkage->oref[i].vn );
    free ( linkage->oref );
  }
  if ( linkage->link ) free ( linkage->link );
  if ( linkage->hj )   free ( linkage->hj );
  if ( linkage->artp ) free ( linkage->artp );
  free ( linkage );
} /*pkkl_DestroyLinkage*/

