
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "pkvaria.h"
#include "pknum.h"

boolean pkn_multiSolveDLSQf ( int nrows, int ncols, const float *a,
                              int spdimen, int bpitch, float *b,
                              int xpitch, const float *xd, float *x )
{
  void  *sp;
  float *at, *aa, *z, *w;

  sp = pkv_GetScratchMemTop ();
  if ( nrows >= ncols )
    goto failure;
  at = pkv_GetScratchMemf ( nrows*(ncols+2)+(ncols+nrows)*spdimen );
  if ( !at )
    goto failure;
  aa = &at[nrows*ncols];
  z = &aa[2*nrows];
  w = &z[nrows*spdimen];
  pkv_TransposeMatrixf ( nrows, ncols, ncols, a, nrows, at );
  pkn_QRDecomposeMatrixf ( ncols, nrows, at, aa );
  pkn_multiMultInvTrUTVectorf ( nrows, at, spdimen, bpitch, b, spdimen, z );
  if ( xd ) {
    pkv_Selectf ( ncols, spdimen, xpitch, spdimen, xd, w );
    pkn_multiReflectVectorf ( ncols, nrows, at, aa, spdimen, spdimen, w );
    pkn_SubtractMatrixf ( 1, nrows*spdimen, 0, z, 0, w, 0, z );
  }
  memset ( &z[nrows*spdimen], 0, (ncols-nrows)*spdimen*sizeof(float) );
  pkn_multiInvReflectVectorf ( ncols, nrows, at, aa, spdimen, spdimen, z );
  pkn_AddMatrixf ( ncols, spdimen, spdimen, z, xpitch, xd, xpitch, x );

  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*pkn_multiSolveDLSQf*/

