
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2016, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <limits.h>
#include <sys/times.h>

#include "pkvaria.h"
#include "avltree.h"
#include "pkscanner.h"

#include "pkbprivate.h"

/* ////////////////////////////////////////////////////////////////////////// */
void pksbatch_ListCompiledCode ( pks_batchexec *batch, FILE *f )
{
  char             *code;
  int              ip, codelgt, nextint;
  double           nextfloat;
  unsigned int     loc;
  short int        token, key, idsp;
  char             *name;
  jumpstatement    *jumpst;
  ifstatement      *ifst;
  whileloop        *whilest;
  forloop          *forst;
  switchstatement  *switchst;
  switchcases      *swcases;
  include_end      *eoi;
  int              i, *casetab;

  if ( !(code = batch->code) )
    return;
  codelgt = batch->codelgt;
  fprintf ( f, "code length: %d\n", codelgt );
  for ( ip = 0; ip < codelgt; ) {
    loc = *(unsigned int*)(&code[ip]);
    if ( loc < UINT_MAX )
      fprintf ( f, "%5d (%4d,%3d): ", ip,
               (loc / PKS_SCANNER_MAX_LINE_LENGTH)+1,
               (loc % PKS_SCANNER_MAX_LINE_LENGTH)+1 );
    else
      fprintf ( f, "%5d (   0,  0): ", ip );
    ip += sizeof(unsigned int);
    token = *(short int*)(&code[ip]);
    ip += sizeof(short int);
    switch ( token ) {
  case PKS_SYMB_ERROR:
      fprintf ( f, "ERROR\n" );
      return;
  case PKS_SYMB_EOLN:
      fprintf ( f, "EOLN\n" );
      break;
  case PKS_SYMB_EOF:
      fprintf ( f, "EOF\n" );
      break;
  case PKS_SYMB_NONE:
      fprintf ( f, "NONE\n" );
      break;
  case PKS_SYMB_PLUS:
      fprintf ( f, "PLUS (+)\n" );
      break;
  case PKS_SYMB_MINUS:
      fprintf ( f, "MINUS (-)\n" );
      break;
  case PKS_SYMB_STAR:
      fprintf ( f, "STAR (*)\n" );
      break;
  case PKS_SYMB_SLASH:
      fprintf ( f, "SLASH (/)\n" );
      break;
  case PKS_SYMB_LPAREN:
      fprintf ( f, "LPAREN (()\n" );
      break;
  case PKS_SYMB_RPAREN:
      fprintf ( f, "RPAREN ())\n" );
      break;
  case PKS_SYMB_LBRACKET:
      fprintf ( f, "LBRACKET ([)\n" );
      break;
  case PKS_SYMB_RBRACKET:
      fprintf ( f, "RBRACKET (])\n" );
      break;
  case PKS_SYMB_LBRACE:
      fprintf ( f, "LBRACE ({)\n" );
      break;
  case PKS_SYMB_RBRACE:
      fprintf ( f, "RBRACE (})\n" );
      break;
  case PKS_SYMB_LANGLE:
      fprintf ( f, "LANGLE (<)\n" );
      break;
  case PKS_SYMB_RANGLE:
      fprintf ( f, "RANGLE (>)\n" );
      break;
  case PKS_SYMB_EQ:
      fprintf ( f, "EQ (=)\n" );
      break;
  case PKS_SYMB_INTEGER:
      nextint = *(int*)(&code[ip]);
      fprintf ( f, "INTEGER (%d)\n", nextint );
      ip += sizeof(int);
      break;
  case PKS_SYMB_FLOAT:
      nextfloat = *(double*)(&code[ip]);
      fprintf ( f, "FLOAT (%f)\n", nextfloat );
      ip += sizeof(double);
      break;
  case PKS_SYMB_CHAR:
      nextint = *(int*)(&code[ip]);
      fprintf ( f, "CHAR (%d,'%c')\n", nextint, nextint );
      ip += sizeof(int);
      break;
  case PKS_SYMB_STRING:
      fprintf ( f, "STRING (\"%s\")\n", &code[ip] );
      ip += strlen( &code[ip] )+1;
      break;
  case PKS_SYMB_COMMA:
      fprintf ( f, "COMMA (,)\n" );
      break;
  case PKS_SYMB_DOT:
      fprintf ( f, "DOT (. ,)\n" );
      break;
  case PKS_SYMB_SEMICOLON:
      fprintf ( f, "SEMICOLON (;)\n" );
      break;
  case PKS_SYMB_COLON:
      fprintf ( f, "COLON (:)\n" );
      break;
  case PKS_SYMB_PERCENT:
      fprintf ( f, "PERCENT (%c)\n", '%' );
      break;
  case PKS_SYMB_HASH:
      fprintf ( f, "HASH (#)\n" );
      break;
  case PKS_SYMB_EXCLAM:
      fprintf ( f, "EXCLAM (!)\n" );
      break;
  case PKS_SYMB_TILDE:
      fprintf ( f, "TILDE (~)\n" );
      break;
  case PKS_SYMB_AT:
      fprintf ( f, "AT (@)\n" );
      break;
  case PKS_SYMB_DOLLAR:
      fprintf ( f, "DOLLAR ($)\n" );
      break;
  case PKS_SYMB_ET:
      fprintf ( f, "ET (&)\n" );
      break;
  case PKS_SYMB_BACKSLASH:
      fprintf ( f, "BACKSLASH (\\)\n" );
      break;
  case PKS_SYMB_QUESTION:
      fprintf ( f, "QUESTION (?)\n" );
      break;
  case PKS_SYMB_VERTBAR:
      fprintf ( f, "VERTBAR (|)\n" );
      break;
  case PKS_SYMB_WEDGE:
      fprintf ( f, "WEDGE (^)\n" );
      break;
  case PKS_SYMB_OTHER:
      nextint = *(int*)(&code[ip]);
      fprintf ( f, "OTHER (%d)\n", nextint );
      ip += sizeof(int);
      break;
  case PKS_SYMB_IDENT:
      idsp = *(short int*)(&code[ip]);  ip += sizeof(short int);
      nextint = *(int*)(&code[ip]);  ip += sizeof(int);
      name = &code[ip];  ip += strlen(name)+1;
      fprintf ( f, "IDENT (%d,%d,%s)\n", idsp, nextint, name );
      break;
  case PKS_SYMB_MACROPAR:
      fprintf ( f, "MACROPAR\n" );
      break;
  case PKS_SYMB_EQEQ:
      fprintf ( f, "EQEQ (==)\n" );
      break;
  case PKS_SYMB_LEQ:
      fprintf ( f, "LEQ (<=)\n" );
      break;
  case PKS_SYMB_GEQ:
      fprintf ( f, "GEQ (>=)\n" );
      break;
  case PKS_SYMB_NEQ:
      fprintf ( f, "NEQ (!=)\n" );
      break;
  case PKS_SYMB_ETET:
      fprintf ( f, "ETET (&&)\n" );
      break;
  case PKS_SYMB_VERTBARBAR:
      fprintf ( f, "VERTBARBAR (||)\n" );
      break;
  case PKS_SYMB_BATCH_CMD:
      key = *(short int*)(&code[ip]);
      ip -= sizeof(short int);
      fprintf ( f, "cmd " );
      switch ( key ) {
    case PKS_BATCHCMD_JUMP:
        jumpst = (jumpstatement*)(&code[ip]);
        fprintf ( f, "JUMP %d\n", jumpst->dest );
        ip += sizeof(jumpstatement);
        break;
    case PKS_BATCHCMD_GOTO:
        jumpst = (jumpstatement*)(&code[ip]);
        fprintf ( f, "GOTO %d\n", jumpst->dest );
        ip += sizeof(jumpstatement);
        break;
    case PKS_BATCHCMD_CONTINUE:
        jumpst = (jumpstatement*)(&code[ip]);
        fprintf ( f, "CONTINUE %d\n", jumpst->dest );
        ip += sizeof(jumpstatement);
        break;
    case PKS_BATCHCMD_IF:
        ifst = (ifstatement*)(&code[ip]);
        fprintf ( f, "IF () %d ; %d\n", ifst->ifptr, ifst->elseptr );
        ip += sizeof(ifstatement);
        break;
    case PKS_BATCHCMD_WHILE:
        whilest = (whileloop*)(&code[ip]);
        fprintf ( f, "WHILE () %d ; %d\n", whilest->loopptr, whilest->endptr );
        ip += sizeof(whileloop);
        break;
    case PKS_BATCHCMD_FOR:
        forst = (forloop*)(&code[ip]);
        fprintf ( f, "FOR ( %d; %d; %d; %d ) %d; %d\n", forst->level,
                     forst->initial, forst->final, forst->incr,
                     forst->loopptr, forst->endptr );
        ip += sizeof(forloop);
        break;
    case PKS_BATCHCMD_SWITCH:
        switchst = (switchstatement*)(&code[ip]);
        fprintf ( f, "SWITCH ( %d )\n", switchst->cases );
        ip += sizeof(switchstatement);
        break;
    case PKS_BATCHCMD_CASES:
        swcases = (switchcases*)(&code[ip]);
        ip += sizeof(switchcases);
        casetab = (int*)(&code[ip]);
        ip += 2*swcases->ncases*sizeof(int);
        fprintf ( f, "SWITCH CASES\n    default: %d\n", swcases->defaultloc );
        for ( i = 0; i < swcases->ncases; i++ )
          fprintf ( f, "    case %d: %d\n", casetab[2*i], casetab[2*i+1] );
        break;
    case PKS_BATCHCMD_INCLUDE:
        ip += sizeof(include_filename);
        name = &batch->code[ip];
        fprintf ( f, "INCLUDE %s\n", name );
        ip += strlen ( name ) + 1;
        break;
    case PKS_BATCHCMD_ENDINCL:
        eoi = (include_end*)&code[ip];
        fprintf ( f, "END INCLUDE, %d\n", eoi->namepos );
        ip += sizeof(include_end);
        break;
    default:
        fprintf ( f, "unknown\n" );
        return;
      }
      break;
  case PKS_SYMB_BATCH_VAR:
      key = *(short int*)(&code[ip]);
      fprintf ( f, "BATCH VAR (%d)\n", key );
      ip += sizeof(short int);
      break;
  default:
      fprintf ( f, "UNKNOWN (%d)\n", token );
      return;
    }
  }
} /*pksbatch_ListCompiledCode*/

