
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2007, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include "pkgeom.h"
#include "multibs.h"

#include "xgedit.h"
#include "xgeprivate.h"

boolean xge_Init ( int argc, char *argv[],
                   int (*callback)(xge_widget*,int,int,short,short),
                   char *title )
{
  if ( !_xge_InitInit ( argc, argv ) )
    return false;
  xge_callback = callback;

  xgevisual = XDefaultVisual ( xgedisplay, xgescreen );
  xge_nplanes = XDisplayPlanes ( xgedisplay, xgescreen );
  xgecolormap = XCreateColormap ( xgedisplay,
                      DefaultRootWindow(xgedisplay),
                      xgevisual, AllocNone );
  if ( !xgecolormap ) {
    printf ( "Error: Cannot create colormap\n" );
    return false;
  }

  _xge_InitWinHints ();
  if ( !title )
    title = xge_p_name;
  xge_NewWindow ( title );
  xge_SetWindow ( 0 );
  _xge_InitGCandCursors ();
  xge_GetWindowSize ();
  _xge_FindAspect ();

  if ( xgevisual->class == TrueColor )
    _xge_MakePalette ();
  else {
    printf ( "Error: True colour visual is needed." );
    return false;
  }
  _xge_FinishInit ();
  return true;
} /*xge_Init*/

void xge_Cleanup ( void )
{
  int i;

#ifdef USE_XEXT_SHAPE
  _xge_DestroySpecialWin ();
#endif
  XFreeGC ( xgedisplay, xgegc );
  for ( i = xge_winnum-1; i >= 0; i-- ) {
    xge_SetWindow ( i );
    XFreePixmap ( xgedisplay, xgepixmap );
    XDestroyWindow ( xgedisplay, xgewindow );
  }
  XCloseDisplay ( xgedisplay );
} /*xge_Cleanup*/

