
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2012, 2016                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include "xgedit.h"
#include "xgeprivate.h"


#define T0 0.5
#define T1 0.53125
#define T2 0.46875
static char xround = 0, yround = 0;

static void _xge_qb_roundpixel ( vector3f *ec, vector3f *v, xpoint *px )
{
  float x, fx, y, fy, t;

  x = ec->x + v->x + 0.5;  fx = floor ( x );
  y = ec->y - v->y + 0.5;  fy = floor ( y );
  switch ( xround ) {
 case 1: t = T1;  break;
 case 2: t = T2;  break;
default: t = T0;  break;
  }
  if ( x-fx < t ) { px->x = (short)fx;      xround = 1; }
             else { px->x = (short)(fx+1);  xround = 2; }
  switch ( yround ) {
 case 1: t = T1;  break;
 case 2: t = T2;  break;
default: t = T0;  break;
  }
  if ( y-fy < t ) { px->y = (short)fy;      yround = 1; }
             else { px->y = (short)(fy+1);  yround = 2; }
} /*_xge_qb_roundpixel*/

static void _xge_r_DrawArc ( vector3f va, vector3f vb, vector3f ec )
{
#define STKSIZE 16

  typedef struct {
      xpoint   px;
      vector3f vx;
      short    level;
    } stackel;

  static float t[9] = {0.7071067812,0.5411961001,0.5097955791,
                       0.5024192862,0.5006029982,0.500150636,
                       0.5000376519,0.5000094125,0.5000023531};
  void     *sp;
  xpoint   last, next;
  float    lastz, nextz;
  short    levela, levelb, levelc;
  vector3f vc;
  stackel  *stack;
  int      stp;

#define NEIGHBOURS(a,b) \
  abs(a.x-b.x) <= 1 && abs(a.y-b.y) <= 1
#define OUTPUT(px,z) \
  { if ( z >= 0.0 ) { \
      if ( _pkv_npix == PKV_BUFSIZE ) { _pkv_npix -= 2;  PKV_FLUSH \
        memcpy ( &_pkv_pixbuf[0], &_pkv_pixbuf[PKV_BUFSIZE-2], 2*sizeof(xpoint) ); \
        _pkv_npix = 2; } \
      if ( _pkv_npix > 1 ) \
        { if ( NEIGHBOURS ( px, _pkv_pixbuf[_pkv_npix-2] ) ) _pkv_npix--; } \
      _pkv_pixbuf[_pkv_npix] = px;  _pkv_npix++; \
    } \
    lastz = z; \
  }
#define PUSH(p,v,l) \
  { if (stp < STKSIZE) { stack[stp].px = p;  stack[stp].vx = (v); \
                         stack[stp].level = l;  stp++; } }
#define READ(p,v,l) \
  { p = stack[stp-1].px;  v = stack[stp-1].vx;  l = stack[stp-1].level; }
#define POP \
  stp--;

  sp = pkv_GetScratchMemTop ();
  stack = pkv_GetScratchMem ( STKSIZE*sizeof(stackel) );
  if ( !stack )
    return;
  stp = 0;  /* stack becomes empty */
  xround = yround = 0;
  _xge_qb_roundpixel ( &ec, &va, &last );
  levela = 0;
  OUTPUT ( last, ec.z+va.z )
  _xge_qb_roundpixel ( &ec, &vb, &next );
  PUSH ( next, vb, 0 )
  do {
    READ ( next, vb, levelb )
    nextz = ec.z+vb.z;
    if ( NEIGHBOURS ( last, next ) && fabs(nextz-lastz) < 0.1 ) {
      POP
      last   = next;
      va     = vb;
      levela = levelb;
      OUTPUT ( last, nextz );
    }
    else {
      levelc = max ( levela, levelb );
      if ( levelc < 9 ) {
        AddVector3f ( &va, &vb, &vc );
        MultVector3f ( t[levelc], &vc, &vc );
      }
      else
        MidPoint3f ( &va, &vb, &vc );
      _xge_qb_roundpixel ( &ec, &vc, &next );
      PUSH ( next, vc, levelc+1 )
    }
  } while ( stp );
  PKV_FLUSH
  pkv_SetScratchMemTop ( sp );
} /*_xge_r_DrawArc*/

static void _xge_QuatRotBallDrawCircle ( short xc, short yc, short r, trans3f *tr,
                                         vector3f *c, vector3f *a, vector3f *b )
{
  vector3f ec, va, vb, uva, uvb;

  TransVector3f ( tr, c, &ec );
  ec.x = (float)xc + (float)r*ec.x;
  ec.y = (float)yc - (float)r*ec.y;
  ec.z *= (float)r;
  TransVector3f ( tr, a, &va );
  MultVector3f ( (float)r, &va, &va );
  SetVector3f ( &uva, -va.x, -va.y, -va.z );
  TransVector3f ( tr, b, &vb );
  MultVector3f ( (float)r, &vb, &vb );
  SetVector3f ( &uvb, -vb.x, -vb.y, -vb.z );
  _xge_r_DrawArc ( va, vb, ec );
  _xge_r_DrawArc ( vb, uva, ec );
  _xge_r_DrawArc ( uva, uvb, ec );
  _xge_r_DrawArc ( uvb, va, ec );
} /*_xge_QuatRotBallDrawCircle*/

void _xge_QuatRotBallDrawCircles ( short xc, short yc, short r, trans3f *tr,
                      void (*outpixels)(const xpoint *buf, int n, void *usrptr),
                      void *usrptr )
{
  static vector3f z = {0.0, 0.0, 0.0},
    e1 = {1.0, 0.0, 0.0}, e2 = {0.0, 1.0, 0.0}, e3 = {0.0, 0.0, 1.0},
    f1 = {0.5*SQRT2, 0.0, 0.0}, f2 = {0.0, 0.5*SQRT2, 0.0},
    g1 = {0.5*SQRT2, 0.5*SQRT2,0.0}, g2 = {0.5*SQRT2, -0.5*SQRT2, 0.0},
    c1 = {0.0, 0.0, 0.5*SQRT2}, c2 = {0.0, 0.0, -0.5*SQRT2};
  void *sp;

  sp = pkv_GetScratchMemTop ();
  _pkv_InitPixelBuffer ();
  if ( _pkv_pixbuf ) {
    _pkv_OutputPixels = outpixels;
    _pkv_pixelusrptr = usrptr;
        /* equator */
    _xge_QuatRotBallDrawCircle ( xc, yc, r, tr, &z, &e1, &e2 );
    _xge_QuatRotBallDrawCircle ( xc, yc, r, tr, &c1, &f1, &f2 );
    _xge_QuatRotBallDrawCircle ( xc, yc, r, tr, &c2, &f1, &f2 );
        /* meridians */
    _xge_QuatRotBallDrawCircle ( xc, yc, r, tr, &z, &e1, &e3 );
    _xge_QuatRotBallDrawCircle ( xc, yc, r, tr, &z, &e2, &e3 );
    _xge_QuatRotBallDrawCircle ( xc, yc, r, tr, &z, &g1, &e3 );
    _xge_QuatRotBallDrawCircle ( xc, yc, r, tr, &z, &g2, &e3 );
  }
  _pkv_DestroyPixelBuffer ();
  pkv_SetScratchMemTop ( sp );
} /*_xge_QuatRotBallDrawCircles*/

/* ////////////////////////////////////////////////////////////////////////// */
/* antialiased image of the quaternion ball widget */
static void _yge_SetPixel1 ( short w, byte *data,
                             short x, short y, byte r, byte g, byte b )
{
  int z;

  z = 3*((int)w*y+x);
  data[z]   = r;
  data[z+1] = g;
  data[z+2] = b;
} /*_yge_SetPixel1*/

static void _yge_SetPixel2 ( short w, byte *data, short x, short y )
{
  int z;

  z = 3*((int)w*(y-1)+x-1);
  data[z] = max ( data[z], 159 );
  data[z+1] = max ( data[z+1], 159 );
  data[z+2] = max ( data[z+2], 159 );
  data[z+3] = data[z+4] = data[z+5] = 255;
  data[z+6] = max ( data[z], 159 );
  data[z+7] = max ( data[z+1], 159 );
  data[z+8] = max ( data[z+2], 159 );
  z += 3*w;
  memset ( &data[z], 255, 9 );
  z += 3*w;
  data[z] = max ( data[z], 159 );
  data[z+1] = max ( data[z+1], 159 );
  data[z+2] = max ( data[z+2], 159 );
  data[z+3] = data[z+4] = data[z+5] = 255;
  data[z+6] = max ( data[z], 159 );
  data[z+7] = max ( data[z+1], 159 );
  data[z+8] = max ( data[z+2], 159 );
} /*_yge_SetPixel2*/

static void _yge_r_DrawArc ( short w, byte *data,
                             vector3f va, vector3f vb, vector3f ec )
{
#define STKSIZE 16

  typedef struct {
      xpoint   px;
      vector3f vx;
      short    level;
    } stackel;

  static float t[9] = {0.7071067812,0.5411961001,0.5097955791,
                       0.5024192862,0.5006029982,0.500150636,
                       0.5000376519,0.5000094125,0.5000023531};
  void     *sp;
  xpoint   last, next;
  float    lastz, nextz;
  short    levela, levelb, levelc;
  vector3f vc;
  stackel  *stack;
  int      stp;

#define _NEIGHBOURS(a,b) \
  abs(a.x-b.x) <= 1 && abs(a.y-b.y) <= 1
#define _OUTPUT(px,z) \
  { if ( z >= 0.0 ) \
      _yge_SetPixel2 ( w, data, px.x, px.y );  \
    lastz = z; \
  }
#define _PUSH(p,v,l) \
  { if (stp < STKSIZE) { stack[stp].px = p;  stack[stp].vx = (v); \
                         stack[stp].level = l;  stp++; } }
#define _READ(p,v,l) \
  { p = stack[stp-1].px;  v = stack[stp-1].vx;  l = stack[stp-1].level; }
#define _POP \
  stp--;

  sp = pkv_GetScratchMemTop ();
  stack = pkv_GetScratchMem ( STKSIZE*sizeof(stackel) );
  if ( !stack )
    return;
  stp = 0;  /* stack becomes empty */
  xround = yround = 0;
  _xge_qb_roundpixel ( &ec, &va, &last );
  levela = 0;
  _OUTPUT ( last, ec.z+va.z )
  _xge_qb_roundpixel ( &ec, &vb, &next );
  _PUSH ( next, vb, 0 )
  do {
    _READ ( next, vb, levelb )
    nextz = ec.z+vb.z;
    if ( _NEIGHBOURS ( last, next ) && fabs(nextz-lastz) < 0.1 ) {
      _POP
      last   = next;
      va     = vb;
      levela = levelb;
      _OUTPUT ( last, nextz );
    }
    else {
      levelc = max ( levela, levelb );
      if ( levelc < 9 ) {
        AddVector3f ( &va, &vb, &vc );
        MultVector3f ( t[levelc], &vc, &vc );
      }
      else
        MidPoint3f ( &va, &vb, &vc );
      _xge_qb_roundpixel ( &ec, &vc, &next );
      _PUSH ( next, vc, levelc+1 )
    }
  } while ( stp );
  pkv_SetScratchMemTop ( sp );
} /*_yge_r_DrawArc*/

static void _yge_QuatRotBallDrawCircle ( short w, byte *data,
                                         short xc, short yc, short r, trans3f *tr,
                                         vector3f *c, vector3f *a, vector3f *b )
{
  vector3f ec, va, vb, uva, uvb;

  TransVector3f ( tr, c, &ec );
  ec.x = (float)xc + (float)r*ec.x;
  ec.y = (float)yc - (float)r*ec.y;
  ec.z *= (float)r;
  TransVector3f ( tr, a, &va );
  MultVector3f ( (float)r, &va, &va );
  SetVector3f ( &uva, -va.x, -va.y, -va.z );
  TransVector3f ( tr, b, &vb );
  MultVector3f ( (float)r, &vb, &vb );
  SetVector3f ( &uvb, -vb.x, -vb.y, -vb.z );
  _yge_r_DrawArc ( w, data, va, vb, ec );
  _yge_r_DrawArc ( w, data, vb, uva, ec );
  _yge_r_DrawArc ( w, data, uva, uvb, ec );
  _yge_r_DrawArc ( w, data, uvb, va, ec );
} /*_yge_QuatRotBallDrawCircle*/

void _yge_QuatRotBallDrawCircles ( short w, byte *data,
                                   short xc, short yc, short r, trans3f *tr )
{
  static vector3f z = {0.0, 0.0, 0.0},
    e1 = {1.0, 0.0, 0.0}, e2 = {0.0, 1.0, 0.0}, e3 = {0.0, 0.0, 1.0},
    f1 = {0.5*SQRT2, 0.0, 0.0}, f2 = {0.0, 0.5*SQRT2, 0.0},
    g1 = {0.5*SQRT2, 0.5*SQRT2,0.0}, g2 = {0.5*SQRT2, -0.5*SQRT2, 0.0},
    c1 = {0.0, 0.0, 0.5*SQRT2}, c2 = {0.0, 0.0, -0.5*SQRT2};
  trans3f idtr;

        /* equator and parallels */
  _yge_QuatRotBallDrawCircle ( w, data, xc, yc, r, tr, &z, &e1, &e2 );
  _yge_QuatRotBallDrawCircle ( w, data, xc, yc, r, tr, &c1, &f1, &f2 );
  _yge_QuatRotBallDrawCircle ( w, data, xc, yc, r, tr, &c2, &f1, &f2 );
        /* meridians */
  _yge_QuatRotBallDrawCircle ( w, data, xc, yc, r, tr, &z, &e1, &e3 );
  _yge_QuatRotBallDrawCircle ( w, data, xc, yc, r, tr, &z, &e2, &e3 );
  _yge_QuatRotBallDrawCircle ( w, data, xc, yc, r, tr, &z, &g1, &e3 );
  _yge_QuatRotBallDrawCircle ( w, data, xc, yc, r, tr, &z, &g2, &e3 );
        /* the grand circle */
  IdentTrans3f ( &idtr );
  _yge_QuatRotBallDrawCircle ( w, data, xc, yc, r, &idtr, &z, &e1, &e2 );
} /*_yge_QuatRotBallDrawCircles*/

boolean _xge_DrawQuaternionBallIcon ( trans3f *tr, boolean active )
{
  void  *sp;
  byte  *data;
  byte  r1, g1, b1, r2, g2, b2;
  short imw, imh, x, y, xc, yc, X;

  sp = pkv_GetScratchMemTop ();
  imw = 3*XGE_QBALLICON_W+4;
  imh = 3*XGE_QBALLICON_H+4;
  data = pkv_GetScratchMem ( 3*imw*imh );
  if ( !data )
    return false;
  xge_GetPixelColour ( xgec_MENU_BACKGROUND, &r1, &g1, &b1 );
  if ( active )
    xge_GetPixelColour ( xgec_QUATROTBALL_A, &r2, &g2, &b2 );
  else
    xge_GetPixelColour ( xgec_QUATROTBALL_B, &r2, &g2, &b2 );
  xc = (imw-1)/2;
  yc = (imh-1)/2;
  for ( y = 0; y < 3; y++ )
    for ( x = 0; x < imw; x++ )  {
      _yge_SetPixel1 ( imw, data, x, y, r1, g1, b1 );
      _yge_SetPixel1 ( imw, data, x, imh-1-y, r1, g1, b1 );
    }
  for ( y = 3; y < yc; y++ ) {
    X = xc-(short)sqrt ( (double)(2*yc*(y-2)+1-y*y) );
    for ( x = 0; x < X; x++ ) {
      _yge_SetPixel1 ( imw, data, x, y, r1, g1, b1 );
      _yge_SetPixel1 ( imw, data, imw-1-x, y, r1, g1, b1 );
      _yge_SetPixel1 ( imw, data, x, imh-1-y, r1, g1, b1 );
      _yge_SetPixel1 ( imw, data, imw-1-x, imh-1-y, r1, g1, b1 );
    }
    for ( x = X; x < imw-1-X; x++ ) {
      _yge_SetPixel1 ( imw, data, x, y, r2, g2, b2 );
      _yge_SetPixel1 ( imw, data, x, imh-1-y, r2, g2, b2 );
    }
  }
  for ( x = 0; x < 2; x++ ) {
    _yge_SetPixel1 ( imw, data, x, yc, r1, g1, b1 );
    _yge_SetPixel1 ( imw, data, imw-1-x, yc, r1, g1, b1 );
  }
  for ( x = 2; x < imw-2; x++ )
    _yge_SetPixel1 ( imw, data, x, yc, r2, g2, b2 );
  _yge_QuatRotBallDrawCircles ( imw, data, xc, yc, xc-3, tr );
  _xge_DrawOnCanvasAA ( XGE_QBALLICON_W, XGE_QBALLICON_H,
                        XGE_QBALLICON_X, XGE_QBALLICON_Y, imw, imh, data );

  pkv_SetScratchMemTop ( sp );
  return true;
} /*_xge_DrawQuaternionBallIcon*/

/* ////////////////////////////////////////////////////////////////////////// */
#ifdef USE_XEXT_SHAPE
void _xge_QuatRotCompSpecialWinSizes ( xge_widget *spqw )
{
#define SPWR(i) xge_specialwin[i].thewinrect
  int          x, y, w, h, r, e, i;
  unsigned int wdt, hgh, b, d;
  Window       root, child;

  XGetGeometry ( xgedisplay, xgewindow, &root, &x, &y, &wdt, &hgh, &b, &d );
        /* make all special windows empty */
  for ( i = 0; i < 4; i++ )
    SPWR(i).width = SPWR(i).height = 0;
  r = (spqw->h+1)/2;
        /* compute the left special window size */
  if ( spqw->x < 0 ) {
    SPWR(0).width = w = -spqw->x;  SPWR(0).height = spqw->h;
    SPWR(0).x = spqw->x;  SPWR(0).y = spqw->y;
    if ( w < r-1 ) {
      e = r - (int)(sqrt ( (double)w*(r+r-w) ) + 1.0);
      SPWR(0).height -= e+e;
      SPWR(0).y += e;
    }
  }
        /* compute the right special window size */
  if ( spqw->x+spqw->w > wdt ) {
    SPWR(1).width = w = spqw->x+spqw->w-wdt;  SPWR(1).height = spqw->h;
    SPWR(1).x = wdt;  SPWR(1).y = spqw->y;
    if ( w < r-1 ) {
      e = r - (int)(sqrt ( (double)w*(r+r-w) ) + 1.0);
      SPWR(1).height -= e+e;
      SPWR(1).y += e;
    }
  }
        /* compute the upper special window size */
  if ( spqw->y < 0 ) {
    SPWR(2).width = spqw->w;  SPWR(2).height = h = -spqw->y;
    SPWR(2).x = spqw->x;  SPWR(2).y = spqw->y;
    if ( h < r-1 ) {
      e = r - (int)(sqrt ( (double)h*(r+r-h) ) + 1.0);
      SPWR(2).width -= e+e;
      SPWR(2).x += e;
    }
    if ( SPWR(2).x < 0 ) { SPWR(2).width += SPWR(2).x;  SPWR(2).x = 0; }
    if ( SPWR(2).x+SPWR(2).width > wdt )
      SPWR(2).width -= SPWR(2).x+SPWR(2).width-wdt;
  }
        /* compute the lower special window size */
  if ( spqw->y+spqw->h > hgh ) {
    SPWR(3).width = spqw->w;  SPWR(3).height = h = spqw->y+spqw->h-hgh;
    SPWR(3).x = spqw->x;  SPWR(3).y = hgh;
    if ( h < r-1 ) {
      e = r - (int)(sqrt ( (double)h*(r+r-h) ) + 1.0);
      SPWR(3).width -= e+e;
      SPWR(3).x += e;
    }
    if ( SPWR(3).x < 0 ) { SPWR(3).width += SPWR(3).x;  SPWR(3).x = 0; }
    if ( SPWR(3).x+SPWR(3).width > wdt )
      SPWR(3).width -= SPWR(3).x+SPWR(3).width-wdt;
  }
        /* do something with it */
  for ( i = 0; i < 4; i++ ) {
    xge_specialwin[i].nonempty = SPWR(i).width > 0 && SPWR(i).height > 0;
    if ( xge_specialwin[i].nonempty ) {
      XTranslateCoordinates ( xgedisplay, xgewindow, root,
                              SPWR(i).x, SPWR(i).y, &x, &y, &child );
      xge_specialwin[i].xpos = x;
      xge_specialwin[i].ypos = y;
      xge_specialwin[i].xpix = spqw->x-SPWR(i).x;
      xge_specialwin[i].ypix = spqw->y-SPWR(i).y;
    }
  }
#undef SPWR
} /*_xge_QuatRotCompSpecialWinSizes*/

void _xge_QuatRotBallDrawSpecialWin ( int w, xge_widget *wdg )
{
  XSetForeground ( xgedisplay, xge_specialwingc,
                   WhitePixel(xgedisplay,DefaultScreen(xgedisplay)) );
  XFillRectangle ( xgedisplay, xge_specialwin[w].thewindow,
                   xge_specialwingc, 0, 0,
                   xge_specialwin[w].thewinrect.width,
                   xge_specialwin[w].thewinrect.height );
} /*_xge_QuatRotBallDrawSpecialWin*/
#endif

