
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <GL/gl.h>
#include <GL/glx.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camera.h"

#define XGLE_OPENGL_NEW
#include "xgledit.h"
#include "xgeprivate.h"

/* ///////////////////////////////////////////////////////////////////////// */
static void xgle_3DwindDrawXYZ ( CameraRecd *CPos, char cc, char left,
                                 int xi, int eta, double xyz )
{
  char s[30];

  xgle_SetForeground ( xgec_Green, 255 );
  xgle_SetBackground ( 0, 0 );
  xgle_Choose12x6font ();
  switch ( cc ) {
case 0: sprintf ( s, "x=%5.3f", xyz );  break;
case 1: sprintf ( s, "y=%5.3f", xyz );  break;
case 2: sprintf ( s, "z=%5.3f", xyz );  break;
  }
  if ( left ) {
    xi = /*CPos->xmin+*/2;
    eta = max ( eta-CPos->ymin, /*CPos->ymin+*/11 );
  }
  else {
    xi = min ( xi-CPos->xmin+2, /*CPos->xmin+*/CPos->width-6*strlen(s) );
    eta = /*CPos->ymin+*/CPos->height-2;
  }
  xgle_DrawString ( s, xi, eta );
} /*xgle_3DwindDrawXYZ*/

void xgle_3DwindDrawCursorPos ( xge_3Dwind *_3Dwin,
                                int id, short x, short y )
{
  point3d p, q;
  char    zoomwin;

  id &= 0x03;
  if ( !_3Dwin->display_coord ||
       _3Dwin->CoordWin < 0 || _3Dwin->CoordWin > 2 ||
       id < 0 || id > 2 )
    return;
  xgle_SetupXGLCoordinates ( _3Dwin->CPos[id].width, _3Dwin->CPos[id].height );
  zoomwin = _3Dwin->fww.zoomwin;
  xgle_SetForeground ( xgec_Green5, 255 );
  switch ( _3Dwin->CoordWin ) {
case 0:
    switch ( id ) {
  case 0:
      SetPoint3d ( &q, (double)x, (double)y, 0.0 );
      xgle_DrawLine ( x-_3Dwin->CPos[id].xmin, 0,
                      x-_3Dwin->CPos[id].xmin, _3Dwin->CPos[id].height );
      xgle_DrawLine ( 0, y-_3Dwin->CPos[id].ymin,
                      _3Dwin->CPos[id].width, y-_3Dwin->CPos[id].ymin ); 
      CameraUnProjectPoint3d ( &_3Dwin->CPos[id], &q, &p );
      xgle_3DwindDrawXYZ ( &_3Dwin->CPos[id], 2, 1, x, y, p.z );
      if ( zoomwin == 0 )
        xgle_3DwindDrawXYZ ( &_3Dwin->CPos[id], 0, 0, x, y, p.x );
      break;
  case 1:   
      xgle_DrawLine ( 0, y-_3Dwin->CPos[id].ymin,
                      _3Dwin->CPos[id].width, y-_3Dwin->CPos[id].ymin );
      break;
  case 2:   
      SetPoint3d ( &q, (double)x, (double)y, 0.0 );
      xgle_DrawLine ( x-_3Dwin->CPos[id].xmin, 0,
                      x-_3Dwin->CPos[id].xmin, _3Dwin->CPos[id].height );
      CameraUnProjectPoint3d ( &_3Dwin->CPos[id], &q, &p );
      xgle_3DwindDrawXYZ ( &_3Dwin->CPos[id], 0, 0, x, y, p.x );
      break;
    }
    break;
case 1:   
    switch ( id ) {
  case 0:
      SetPoint3d ( &q, (double)x, (double)y, 0.0 );
      xgle_DrawLine ( 0, y-_3Dwin->CPos[id].ymin,
                      _3Dwin->CPos[id].width, y-_3Dwin->CPos[id].ymin );
      CameraUnProjectPoint3d ( &_3Dwin->CPos[id], &q, &p );
      xgle_3DwindDrawXYZ ( &_3Dwin->CPos[id], 2, 1, x, y, p.z );
      break;
  case 1:   
      SetPoint3d ( &q, (double)x, (double)y, 0.0 );
      xgle_DrawLine ( x-_3Dwin->CPos[id].xmin, 0,
                      x-_3Dwin->CPos[id].xmin, _3Dwin->CPos[id].height );
      xgle_DrawLine ( 0, y-_3Dwin->CPos[id].ymin,
                      _3Dwin->CPos[id].width, y-_3Dwin->CPos[id].ymin ); 
      CameraUnProjectPoint3d ( &_3Dwin->CPos[id], &q, &p );
      xgle_3DwindDrawXYZ ( &_3Dwin->CPos[id], 1, 0, x, y, p.y );
      if ( zoomwin == 1 )
        xgle_3DwindDrawXYZ ( &_3Dwin->CPos[id], 2, 1, x, y, p.z );
      break;
  case 2:   
      SetPoint3d ( &q, (double)x, (double)y, 0.0 );
      CameraUnProjectPoint3d ( &_3Dwin->CPos[_3Dwin->CoordWin], &q, &p );
      CameraProjectPoint3d ( &_3Dwin->CPos[id], &p, &q );   
      xgle_DrawLine ( 0, (int)q.y-_3Dwin->CPos[id].ymin,
                      _3Dwin->CPos[id].width, (int)q.y-_3Dwin->CPos[id].ymin );
      break;
    }
    break;
case 2:   
    switch ( id ) {
  case 0:
      xgle_DrawLine ( x-_3Dwin->CPos[id].xmin, 0,
                      x-_3Dwin->CPos[id].xmin, _3Dwin->CPos[id].height );
      break;
  case 1:   
      SetPoint3d ( &q, (double)x, (double)y, 0.0 );
      CameraUnProjectPoint3d ( &_3Dwin->CPos[_3Dwin->CoordWin], &q, &p );
      CameraProjectPoint3d ( &_3Dwin->CPos[id], &p, &q );   
      xgle_DrawLine ( (int)q.x-_3Dwin->CPos[id].xmin, 0,
                      (int)q.x-_3Dwin->CPos[id].xmin, _3Dwin->CPos[id].height );
      break;
  case 2:   
      SetPoint3d ( &q, (double)x, (double)y, 0.0 );
      xgle_DrawLine ( x-_3Dwin->CPos[id].xmin, 0,
                      x-_3Dwin->CPos[id].xmin, _3Dwin->CPos[id].height );
      xgle_DrawLine ( 0, y-_3Dwin->CPos[id].ymin,
                      _3Dwin->CPos[id].width, y-_3Dwin->CPos[id].ymin ); 
      CameraUnProjectPoint3d ( &_3Dwin->CPos[id], &q, &p );
      xgle_3DwindDrawXYZ ( &_3Dwin->CPos[id], 1, 1, x, y, p.y );
      xgle_3DwindDrawXYZ ( &_3Dwin->CPos[id], 0, 0, x, y, p.x );
      break;
    }
    break;
  }
} /*xgle_3DwindDrawCursorPos*/

