
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include <GL/gl.h>
#include <GL/glx.h>

#include "pkgeom.h"
#include "camera.h"

#define XGLE_OPENGL_NEW
#include "xgledit.h"
#include "xgeprivate.h"

/* ///////////////////////////////////////////////////////////////////////// */
/* This procedure computes the GL projection/modelview matrices as described */
/* by the CameraRec structure, processed by the libcamera procedures         */

boolean xgle_SetGLCamerad ( CameraRecd *CPos, GLint viewport[4],
                            GLdouble view[16], GLdouble projection[16],
                            GLdouble eyepos[4] )
{
  double  w, h, ac, shx, shy, s;
  trans3d vt;

  if ( CPos->ncplanes < 6 || CPos->zmax <= CPos->zmin )
    return false;
  if ( !CPos->parallel && CPos->zmin <= 0.0 )
    return false;

  viewport[0] = viewport[1] = 0;
  viewport[2] = CPos->width/CPos->magnification;
  viewport[3] = CPos->height/CPos->magnification;

  IdentTrans3d ( &vt );
  ShiftTrans3d ( &vt, -CPos->position.x, -CPos->position.y, -CPos->position.z );
  EulerRotTrans3d ( &vt, -CPos->psi, -CPos->theta, -CPos->phi );
  ScaleTrans3d ( &vt, 1.0, 1.0, -1.0 );
  xgle_Trans3toGLd ( &vt, view );  

  if ( CPos->parallel ) {
    w = CPos->vd.para.wdt;
    h = CPos->vd.para.hgh;
    if ( CPos->upside )
      xgle_Orthod ( -0.5*w, 0.5*w, -0.5*h, 0.5*h, CPos->zmin, CPos->zmax,
                    projection, NULL );
    else
      xgle_Orthod ( -0.5*w, 0.5*w, 0.5*h, -0.5*h, CPos->zmin, CPos->zmax,
                    projection, NULL );
    IdentTrans3d ( &vt );
    EulerRotTrans3d ( &vt, CPos->phi, CPos->theta, CPos->phi );
    eyepos[0] = -vt.U0.a13;
    eyepos[1] = -vt.U0.a23;
    eyepos[2] = -vt.U0.a33;
    eyepos[3] = 0.0;
  }
  else {
    ac = (double)CPos->height/(CPos->aspect*(double)CPos->width);
    w = sqrt ( 1.0/(1.0+ac*ac) );
    h = ac*w;
    shx = CPos->vd.persp.dxi0/CPos->xscale;
    shy = CPos->vd.persp.deta0/CPos->yscale;
    s = CPos->zmin/CPos->vd.persp.f;
    if ( CPos->upside )
      xgle_Frustumd ( s*(shx-0.5*w), s*(shx+0.5*w), s*(shy-0.5*h), s*(shy+0.5*h),
                      CPos->zmin, CPos->zmax, projection, NULL );
    else
      xgle_Frustumd ( s*(shx-0.5*w), s*(shx+0.5*w), s*(shy+0.5*h), s*(shy-0.5*h),
                      CPos->zmin, CPos->zmax, projection, NULL );
    eyepos[0] = CPos->position.x;
    eyepos[1] = CPos->position.y;
    eyepos[2] = CPos->position.z;
    eyepos[3] = 1.0;
  }
  return true;
} /*xgle_SetGLCamerad*/

