
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "pkvaria.h"
#include "psout.h"

#include "avltree.h"

/* ////////////////////////////////////////////////////////////////////////// */
typedef struct {
    AVLtnode node;
    int      key;
    int      w, x, y;
  } tnode;

char compkeys ( AVLtnode *node, void *data )
{
  tnode *tn;
  int   *mydata;

  tn = (tnode*)node;
  mydata = (int*)data;
  if ( tn->key < *mydata )      return -1;
  else if ( tn->key > *mydata ) return +1;
  else                          return  0;
} /*compkeys*/

AVLtnode *allocnode ( void *data )
{
  tnode *tn;

  tn = malloc ( sizeof(tnode) );
  if ( tn )
    tn->key = *((int*)data);
  return &tn->node;
} /*allocnode*/

void freenode ( AVLtnode *node )
{
  free ( node );
} /*freenode*/

/* ////////////////////////////////////////////////////////////////////////// */

#define VWD    250
#define VHG    150
#define VHDIST  20
#define VVDIST 100
#define XSH     10
#define YSH     40

void AVLSetW ( tnode *p )
{
  if ( p->node.left )
    AVLSetW ( (tnode*)p->node.left );
  if ( p->node.right )
    AVLSetW ( (tnode*)p->node.right );
  if ( p->node.left && p->node.right )
    p->w = ((tnode*)(p->node.left))->w+((tnode*)(p->node.right))->w+VHDIST;
  else if ( !p->node.left && !p->node.right )
    p->w = VWD;
  else if ( p->node.left )
    p->w = ((tnode*)(p->node.left))->w+VWD/2+VHDIST;
  else p->w = ((tnode*)(p->node.right))->w+VWD/2+VHDIST;
} /*AVLSetW*/

void AVLSetXY ( tnode *p, int x, int y )
{
  p->y = y;
  y -= VVDIST+VHG;
  if ( p->node.left ) {
    AVLSetXY ( (tnode*)p->node.left, x, y );
    if ( p->node.right ) {
      AVLSetXY ( (tnode*)p->node.right, x+((tnode*)(p->node.left))->w+VHDIST, y );
      p->x = (((tnode*)(p->node.left))->x+((tnode*)(p->node.right))->x)/2;
    }
    else
      p->x = ((tnode*)(p->node.left))->x+VWD/2+VHDIST;
  }
  else {
    if ( p->node.right ) {
      AVLSetXY ( (tnode*)p->node.right, x+VWD/2+VHDIST, y );
      p->x = ((tnode*)(p->node.right))->x-VWD/2-VHDIST;
    }
    else
      p->x = x;
  }
} /*AVLSetXY*/

void DrawArrowLine ( int x0, int y0, int x1, int y1 )
{
  point2d cp[4];

  SetPoint2d ( &cp[0], x0, y0 );
  SetPoint2d ( &cp[1], 0.9*x0+0.1*x1, 0.4*y0+0.6*y1 );
  SetPoint2d ( &cp[2], x1, 0.9*y0+0.1*y1 );
  SetPoint2d ( &cp[3], x1, 0.25*y0+0.75*y1 );
  ps_Set_Line_Width ( 2.0 );
  ps_Draw_BezierCd ( cp, 3 );
  psl_SetLine ( cp[2].x, cp[2].y, x1, y1, 0.0, 1.0 );
  ps_Draw_Line ( cp[3].x, cp[3].y, x1, y1 );
  psl_MapsTo ( 1.0 );
} /*DrawArrowLine*/

void AVLDrawPicture ( tnode *p )
{
  char s[60];

  ps_Set_Line_Width ( 3.0 );
  ps_Draw_Rect ( VWD, VHG, p->x, p->y );
/*  sprintf ( s, "%d %d moveto (%d:%d) show",
              p->x+XSH, p->y+YSH, p->klucz, p->dana ); */
  sprintf ( s, "a %d %d moveto (%d) show", p->x+XSH, p->y+YSH, p->key );
  ps_Write_Command ( s );
  ps_Set_RGB ( 0.0, 0.0, 1.0 );
  sprintf ( s, "b %d %d moveto (%d) show", p->x+VWD-70, p->y+YSH, p->node.eqi );
  ps_Write_Command ( s );
  ps_Set_Gray ( 0.0 );
  if ( p->node.left ) {
    DrawArrowLine ( p->x+VWD/4, p->y,
                    ((tnode*)(p->node.left))->x+VWD/2,
                    ((tnode*)(p->node.left))->y+VHG );
    AVLDrawPicture ( (tnode*)p->node.left );
  }
  if ( p->node.right ) {
    DrawArrowLine ( p->x+3*VWD/4, p->y,
                    ((tnode*)(p->node.right))->x+VWD/2,
                    ((tnode*)(p->node.right))->y+VHG );
    AVLDrawPicture ( (tnode*)p->node.right );
  }
} /*AVLDrawPicture*/

void AVLDrawTree ( tnode *root )
{
  if ( root ) {
    AVLSetW ( root );
    AVLSetXY ( root, 20, 3000 );
    AVLDrawPicture ( root );
  }
} /*AVLDrawTree*/

/* ////////////////////////////////////////////////////////////////////////// */
int main ( void )
{
#define NK 28
  int     k[NK] = {7,17,9,19,10,20,14,4,13,3,
                   2,12,6,16,5,15,1,11,10,0,
                   8,18,11,21,22,23,24,25};
  int     i;
  AVLtree tree;

  pkv_InitScratchMem ( 1048576 );
  AVLInit ( &tree, true, compkeys, allocnode, freenode );
  ps_OpenFile ( "AVLtree.ps", 600 );
  ps_Write_Command ( "/a {/Times-Roman findfont 100 scalefont setfont} def" );
  ps_Write_Command ( "/b {/Times-Roman findfont 70 scalefont setfont} def" );
  for ( i = 0; i < NK; i++ )
    AVLInsert ( &tree, &k[i] );
  ps_GSave ();
  ps_Write_Command ( "0 3000 translate" );
  AVLDrawTree ( (tnode*)tree.root );
  ps_GRestore ();
  i = 4;
  AVLDelete ( &tree, &i );
  AVLDrawTree ( (tnode*)tree.root );
  ps_CloseFile ();
  pkv_DestroyScratchMem ();
  exit ( 0 );
} /*main*/

