
#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camera.h"
#include "multibs.h"
#include "raybez.h"
#include "psout.h"


#define MAXLEVEL  16
#define MAXINTERS 10

char fn[] = "rayrtrimbsd.ps";

/* ////////////////////////////////////////////////////////////////////////// */
#define degu 3
#define LKNU 10
#define degv 3
#define LKNV 10
double knots_u[] = {-0.5, 0.0, 0.0, 0.0, 1.0, 2.0, 3.0, 4.0, 4.0, 4.0, 4.5};
double knots_v[] = {-0.5, 0.0, 0.0, 0.0, 1.0, 2.5, 2.5, 4.0, 4.0, 4.0, 4.5};
point4d cpoints[] =
  {{-3.0,-3.0,-0.5,1.0}, {-2.7,-3.0,-0.5,1.0}, {-1.5,-3.0,-0.5,1.0},
   {0.0,-3.0,-0.5,1.0},  {1.5,-3.0,-0.5,1.0}, {2.7,-3.0,-0.5,1.0},
   {3.0,-3.0,-0.5,1.0},
   {-3.0,-2.7,0.0,1.0}, {-2.7,-2.7,0.0,1.0}, {-1.5,-2.7,0.0,1.0},
   {0.0,-2.7,0.0,1.0}, {1.5,-2.7,0.0,1.0}, {2.7,-2.7,0.0,1.0},
   {3.0,-2.7,0.0,1.0},
   {-3.0,-1.5,0.1,1.0}, {-2.7,-1.5,0.1,1.0}, {-1.5,-1.5,0.1,1.0},
   {0.0,-1.5,0.1,1.0}, {3.0,-3.0,5.0,2.0}, {2.7,-1.5,0.1,1.0},
   {3.0,-1.5,0.1,1.0},
   {-3.0,0.0,0.2,1.0}, {-2.7,0.0,0.2,1.0}, {-1.5,0.0,0.2,1.0},
   {0.0,0.0,0.2,1.0}, {1.5,0.0,0.2,1.0}, {2.7,0.0,0.2,1.0},
   {3.0,0.0,0.2,1.0},
   {-3.0,1.5,0.1,1.0}, {-2.7,1.5,0.1,1.0}, {-1.5,1.5,0.1,1.0},
   {0.0,1.5,0.1,1.0}, {1.5,1.5,0.1,1.0}, {2.7,1.5,0.1,1.0},
   {3.0,1.5,0.1,1.0},
   {-3.0,2.7,0.0,1.0}, {-2.7,2.7,0.0,1.0}, {-1.5,2.7,0.0,1.0},
   {0.0,2.7,0.0,1.0}, {1.5,2.7,0.0,1.0}, {2.7,2.7,0.0,1.0},
   {3.0,2.7,0.0,1.0},
   {-3.0,3.0,-0.5,1.0}, {-2.7,3.0,-0.5,1.0}, {-1.5,3.0,-0.5,1.0},
   {0.0,3.0,-0.5,1.0}, {1.5,3.0,-0.5,1.0}, {2.7,3.0,-0.5,1.0},
   {3.0,3.0,-0.5,1.0}};

#define  n1   3
#define  NN1 10
double u1[NN1+1] = {-0.5, 0.0, 0.0, 0.0, 1.0, 2.0, 3.0, 4.0, 4.0, 4.0, 4.5};
#define  m1   3   
#define  MM1 10
double v1[MM1+1] = {-0.5, 0.0, 0.0, 0.0, 1.0, 2.5, 2.5, 4.0, 4.0, 4.0, 4.5};

#define nt1a  3
#define NNt1a 10
double ut1a[NNt1a+1] =
  {-0.5, 0.0, 0.0, 0.0, 1.4, 2.8, 4.2, 5.6, 5.6, 5.6, 6.1};
point2d cpt1a[NNt1a-nt1a] =
  {{4.0,0.4},{3.5,0.4},{2.8,0.8},{2.6,2.0},{2.8,3.2},{3.5,3.6},{4.0,3.6}};
point2d cpd1a[3] = {{4.0,3.6},{4.0,4.0},{1.2,4.0}};
#define nt1b  3
#define NNt1b 7
point3d cpt1b[nt1b+1] =
  {{1.2,4.0,1.0},{1.5,3.0,1.0},{0.5,2.5,0.75},{0.0,2.5,1.0}};
point2d cpd1b[4] = {{0.0,2.5},{0.0,0.0},{4.0,0.0},{4.0,0.4}};
point2d cpd1c[5] = {{0.3,1.1},{1.6,1.6},{2.1,1.1},{1.6,0.6},{0.3,1.1}};
#define nt1c 3
point3d cpt1c[4] = {{2.0,3.3,1.0},{0.6,1.65,0.5},{0.6,1.25,0.5},{2.0,2.5,1.0}};
#define nt1d 3
point3d cpt1d[4] = {{2.0,2.5,1.0},{1.25,1.25,0.5},{1.25,1.5,0.5},{2.0,3.0,1.0}};
point2d cpd1e[3] = {{2.0,3.0},{2.0,3.9},{2.0,3.3}};
point2d cpd1f[5] = {{2.0,2.2},{2.3,2.0},{2.0,1.7},{2.0,1.9},{2.0,2.2}};
mbs_polycurved boundary1[9] =
  {{0,0,2,nt1a,NNt1a,&ut1a[0],(double*)&cpt1a[0]},   /* B-spline curve */
   {1,0,2,   1,    2,    NULL,(double*)&cpd1a[0]},   /* polyline */ 
   {2,0,3,nt1b,   -1,    NULL,(double*)&cpt1b[0]},   /* Bezier curve */  
   {3,1, 2,   1,    3,    NULL,(double*)&cpd1b[0]},  /* polyline */
   {4,2, 2,   1,    4,    NULL,(double*)&cpd1c[0]},  /* polyline */
   {5,0,3,nt1c,   -1,    NULL,(double*)&cpt1c[0]},   /* Bezier curve */
   {6,0,3,nt1d,   -1,    NULL,(double*)&cpt1d[0]},   /* Bezier curve */
   {7,1, 2,   1,    2,    NULL,(double*)&cpd1e[0]},  /* line */
   {8,2, 2,   1,    4,    NULL,(double*)&cpd1f[0]}}; /* polyline */


#define WIDTH  1280
#define HEIGHT  960
#define SCF       0.5

CameraRecd CPos;
vector3d   ld;

RBezPatchTreedp tree;

point2d frame[3] = {{1300.0,2.0},{240.0,0.0},{0.0,240.0}};

/* ////////////////////////////////////////////////////////////////////////// */
static void MapPoint2 ( const point2d *frame, const point2d *p, point2d *q )
{
  point2d r;

  r = *p;
  AddVector2Md ( &frame[0], &frame[1], r.x, q );
  AddVector2Md ( q, &frame[2], r.y, q );
} /*MapPoint2*/

void DrawPolycurve ( int nel, mbs_polycurved *poly )
{
#define DD 20
  void    *sp;
  int     i, j, k;
  int     dim, deg, lkn, ku;
  point2d *cc, q, r;
  double  *bsbuf;

  sp = pkv_GetScratchMemTop ();
  cc = pkv_GetScratchMem ( (DD+1)*sizeof(point2d) );
  ps_Set_Gray ( 0.0 );
  for ( i = 0; i < nel; i++ ) {
    if ( poly[i].knots ) {
      dim = poly[i].cpdimen;
      deg = poly[i].degree;
      lkn = poly[i].lastknot;
      ku = mbs_NumKnotIntervalsd ( deg, lkn, poly[i].knots );
      bsbuf = pkv_GetScratchMemd ( ku*dim*(deg+1) );
      if ( !bsbuf )
        exit ( 1 );
      if ( !mbs_multiBSCurvesToBezd ( dim, 1, deg, lkn, poly[i].knots,
                                      0, poly[i].points, &ku,
                                      NULL, NULL, 0, bsbuf ) )
        exit ( 1 );
      if ( dim == 2 ) {
        for ( j = 0; j < ku; j++ ) {
          for ( k = 0; k <= DD; k++ )
            mbs_BCHornerC2d ( deg, &bsbuf[j*2*(deg+1)],
                              (double)k/(double)DD, &cc[k] );
          for ( k = 0; k <= DD; k++ )
            MapPoint2 ( frame, &cc[k], &cc[k] );
          ps_Draw_Polyline2d ( DD+1, cc );
          psl_SetLine ( cc[DD-1].x, cc[DD-1].y, cc[DD].x, cc[DD].y, 0.0, 1.0 );
/*          psl_SmallArrow ( 1.0, true );*/
        }
      }
      else {
        for ( j = 0; j < ku; j++ ) {
          for ( k = 0; k <= DD; k++ )
            mbs_BCHornerC2Rd ( deg, (point3d*)&bsbuf[j*3*(deg+1)],
                               (double)k/(double)DD, &cc[k] );
          for ( k = 0; k <= DD; k++ )
            MapPoint2 ( frame, &cc[k], &cc[k] );
          ps_Draw_Polyline2d ( DD+1, cc );
          psl_SetLine ( cc[DD-1].x, cc[DD-1].y, cc[DD].x, cc[DD].y, 0.0, 1.0 );
/*          psl_SmallArrow ( 1.0, true );*/
        }
      }
      pkv_SetScratchMemTop ( bsbuf );
    }
    else if ( poly[i].lastknot == -1 ) {  /* a Bezier curve */
      if ( poly[i].cpdimen == 2 ) {
        for ( j = 0; j <= DD; j++ )
          mbs_BCHornerC2d ( poly[i].degree, poly[i].points,
                            (double)j/(double)DD, &cc[j] );
      }
      else {
        for ( j = 0; j <= DD; j++ )
          mbs_BCHornerC2Rd ( poly[i].degree, (point3d*)poly[i].points,
                              (double)j/(double)DD, &cc[j] );
      }
      for ( j = 0; j <= DD; j++ )
        MapPoint2 ( frame, &cc[j], &cc[j] );
      ps_Draw_Polyline2d ( DD+1, cc );
      psl_SetLine ( cc[DD-1].x, cc[DD-1].y, cc[DD].x, cc[DD].y, 0.0, 1.0 );
/*      psl_SmallArrow ( 1.0, true );*/
    }
    else { /* a polyline */
      if ( poly[i].cpdimen == 2 ) {
        for ( j = 0; j < poly[i].lastknot; j++ ) {
          MapPoint2 ( frame, (point2d*)&poly[i].points[2*j], &q );
          MapPoint2 ( frame, (point2d*)&poly[i].points[2*j+2], &r );
          ps_Draw_Line ( r.x, r.y, q.x, q.y );
          psl_SetLine ( q.x, q.y, r.x, r.y, 0.0, 1.0 );
/*          psl_SmallArrow ( 1.0, true );*/
        }
      }
      else {
        SetPoint2d ( &r, poly[i].points[0]/poly[i].points[2],
                         poly[i].points[1]/poly[i].points[2] );
        MapPoint2 ( frame, &r, &r );
        for ( j = 1; j <= poly[i].lastknot; j++ ) {
          SetPoint2d ( &q, poly[i].points[3*j+0]/poly[i].points[3*j+2],
                           poly[i].points[3*j+1]/poly[i].points[3*j+2] );
          MapPoint2 ( frame, &q, &q );
          ps_Draw_Line ( r.x, r.y, q.x, q.y );
          psl_SetLine ( r.x, r.y, q.x, q.y, 0.0, 1.0 );
/*          psl_SmallArrow ( 1.0, true );*/
          r = q;
        }
      }
    }
  }
  pkv_SetScratchMemTop ( sp );
#undef DD
} /*DrawPolycurve*/

void r_DrawBSTree ( RBezPatchTreeVertexdp vertex )
{
  point2d p0, p1, q0, q1;

  if ( !vertex )
    return;
  if ( vertex->left ) {
    r_DrawBSTree ( vertex->left );
    r_DrawBSTree ( vertex->right );
    if ( vertex->divdir == 0 )
      SetPoint2d ( &p0, vertex->left->u1, vertex->left->v0 );
    else
      SetPoint2d ( &p0, vertex->left->u0, vertex->left->v1 );
    SetPoint2d ( &p1, vertex->left->u1, vertex->left->v1 );
    MapPoint2 ( frame, &p0, &q0 );
    MapPoint2 ( frame, &p1, &q1 );
    ps_Set_Line_Width ( 1.0 );
    ps_Set_Gray ( 0.0 );
    ps_Draw_Line ( q0.x, q0.y, q1.x, q1.y );
  }
  else {
    switch ( vertex->colour ) {
  case RAYBEZ_WHITE:  return;
  case RAYBEZ_BLACK:  ps_Set_Gray ( 0.25 );  break;
  default: ps_Set_RGB ( 0.65, 0.75, 1.0 );   break;
    }
    SetPoint2d ( &p0, vertex->u0, vertex->v0 );  MapPoint2 ( frame, &p0, &q0 );
    SetPoint2d ( &p1, vertex->u1, vertex->v1 );  MapPoint2 ( frame, &p1, &q1 );
    ps_Fill_Rect ( q1.x-q0.x, q1.y-q0.y, q0.x, q0.y );
    if ( vertex->trd )
      DrawPolycurve ( vertex->ntrd, vertex->trd );
  }
} /*r_DrawBSTree*/

void DrawBSTree ( RBezPatchTreedp tree )
{
  point2d               p0, p1, q0, q1;
  RBezPatchTreeVertexdp root;

  root = tree->root;
  r_DrawBSTree ( root );
  SetPoint2d ( &p0, root->u0, root->v0 );  MapPoint2 ( frame, &p0, &q0 );
  SetPoint2d ( &p1, root->u1, root->v1 );  MapPoint2 ( frame, &p1, &q1 );
  ps_Set_Gray ( 0.0 );
  ps_Set_Line_Width ( 2.0 );
  ps_Draw_Rect ( q1.x-q0.x, q1.y-q0.y, q0.x, q0.y );
} /*DrawBSTree*/

/* ////////////////////////////////////////////////////////////////////////// */
static void SetCPos ( void )
{
  CameraInitFramed ( &CPos, false, false, WIDTH, HEIGHT, 0, 0, 1.0, 4 );
  CameraInitPosd ( &CPos );
  SetPoint3d ( &CPos.position, -15.31251039605, -72.93132382473, 49.231894764605 );
  CPos.psi   = 2.963570202665;
  CPos.theta = 2.1560152051;
  CPos.phi   = 2.934640744727;
  CPos.vd.persp.f = 10.0;
  CameraSetMappingd ( &CPos );
  CameraRotZGd ( &CPos, 0.5*PI );
} /*SetCPos*/

static void SetLightDir ( void )
{
  SetVector3d ( &ld, 0.5, 0.5, -1.0 );
  NormalizeVector3d ( &ld );
} /*SetLightDir*/

static byte PixelColour ( ray3d *ray, void *workspace )
{
  RayObjectIntersd inters[MAXINTERS];
  int ninters, i, j;
  double tmin;
  double gray, g0, l0;

  _rbez_FindRayRBezPatchIntersd ( tree, ray, MAXLEVEL, MAXINTERS,
                                  &ninters, inters, workspace );
  if ( ninters ) {
    if ( ninters > 1 ) {
      j = 0;
      tmin = inters[0].t;
      for ( i = 0; i < ninters; i++ )
        if ( inters[i].t < tmin ) {
          tmin = inters[i].t;
          j = i;
        }
      if ( j )
        inters[0] = inters[j];
    }
    gray = 0.2;
    l0 = DotProduct3d ( &ray->v, &inters[0].nv );
    g0 = DotProduct3d ( &ld, &inters[0].nv );
    if ( g0*l0 > 0.0 ) {
      gray += 0.7*fabs ( g0 );
      if ( gray > 1.0 )
        gray = 1.0;
    }
    return (byte)(255.0*gray);
  }
  else
    return 160;
} /*PixelColour*/

void RayTrace0 ( int width, int height, byte *image )
{
  void  *sp;
  int   x, y;
  ray3d ray;
  byte  *buf;

  sp = pkv_GetScratchMem ( rbez_RayRBezpWspSized ( degu, degv, MAXLEVEL,
                                                   tree->maxtrdeg ) );
  if ( !sp )
    exit ( 1 );
  for ( y = 0; y < height; y++ ) {
    printf ( "%4d\b\b\b\b", y );
    buf = &image[y*width];
    for ( x = 0; x < width; x++ ) {
      CameraRayOfPixeld ( &CPos, x, y, &ray );
      buf[x] = PixelColour ( &ray, sp );
    }
  }
  pkv_SetScratchMemTop ( sp );
} /*RayTrace0*/

void r_CountTreeVertices ( RBezPatchTreeVertexd *vertex, int *nv )
{
  int k;

  if ( !vertex )  
    return;
  
  switch ( vertex->colour ) {
case RAYBEZ_WHITE: k = 0;  break;
case RAYBEZ_GREY : k = 1;  break;
case RAYBEZ_BLACK: k = 2;  break;
default: exit ( 1 );
  }
  if ( vertex->tag != 2 )
    k += 3;
  nv[k] ++;
  r_CountTreeVertices ( vertex->left, nv );
  r_CountTreeVertices ( vertex->right, nv );
} /*r_CountTreeVertices*/

void CountTreeVertices ( RBezPatchTreed *tree, int *nv )
{
  memset ( nv, 0, 6*sizeof(int) );
  r_CountTreeVertices ( tree->root, nv );
} /*CountTreeVertices*/

void test ( void )
{
  double     *workspace;
  byte       *image;
  char       s[32];
  int        y;
  struct tms start, stop;
  clock_t    ta, tb;
  double     time;
  int        nv[6];
  
  image = malloc ( WIDTH*HEIGHT*sizeof(byte) );
  if ( !image )
    exit ( 1 );
  workspace = pkv_GetScratchMemd ( 2048 /*34*4*/ );
  tree = rbez_NewTrimmedRBSPatchTreed ( 0, degu, LKNU, knots_u, degv, LKNV, knots_v,
                                        LKNV-degv, cpoints, 9, boundary1,
                                        workspace );
  if ( tree ) {
    ps_WriteBBox ( -1, -1, 204, 88 );
    ps_OpenFile ( fn, 400 );
    ps_Write_Command ( "1 setlinecap" );
    SetCPos ();
    SetLightDir ();
    ta = times ( &start );
    RayTrace0 ( WIDTH, HEIGHT, image );
    tb = times ( &stop );
    time = (double)(tb-ta)/(double)(sysconf(_SC_CLK_TCK));
    printf ( "time = %8.3f\n", time );
    ps_GSave ();
    sprintf ( s, "%4.2f dup scale", SCF );
    ps_Write_Command ( s );
    ps_Init_BitmapP ( WIDTH, HEIGHT, 0, 0 );
    for ( y = 0; y < HEIGHT; y++ ) {
      printf ( "%4d\b\b\b\b", y );
      ps_Out_LineP ( &image[y*WIDTH] );
    }
    printf ( "\n" );
    DrawBSTree ( tree );
    CountTreeVertices ( tree, nv );
    rbez_DestroyRBezPatchTreed ( tree );
    ps_CloseFile ();
    printf ( "%s\n", fn );
    printf ( "inner  WHITE:%4d, GREY :%4d, BLACK:%4d\n", nv[0], nv[1], nv[2] );
    printf ( "leaves WHITE:%4d, GREY :%4d, BLACK:%4d\n", nv[3], nv[4], nv[5] );
  }
  pkv_SetScratchMemTop ( workspace );
  free ( image );
} /*test*/

int main ( int argc, char **argv )
{
  setvbuf ( stdout, NULL, _IONBF, 0 );
  pkv_InitScratchMem ( 10485760 );
  test ();
  pkv_DestroyScratchMem ();
  exit ( 0 );
} /*main*/

